"""TcEx Framework Module"""

# standard library
import hashlib
import json
import uuid
from collections.abc import Iterable
from inspect import _empty, signature
from typing import TypedDict

# first-party
from tcex.api.tc.ti_transform.model.transform_model import (
    GroupTransformModel,
    IndicatorTransformModel,
)


class TransformBuilderExport(TypedDict):
    """Basic definition of a transform exported from Transform Builder."""

    type: str
    transform: dict


def transform_builder_to_model(
    transform: TransformBuilderExport,
    processing_functions: 'ProcessingFunctions',
) -> IndicatorTransformModel | GroupTransformModel:
    """Convert a transform from Transform Builder to one of the tcex transform models."""

    def find_entries(data, key, context='') -> Iterable[tuple[str, dict]]:
        """Find entries in a dict with a given name, regardless of depth."""
        if isinstance(data, dict):
            for k, v in data.items():
                if k == key:
                    yield (context, v)
                elif isinstance(v, dict | list):
                    yield from find_entries(v, key, context=f'{context}.{k}' if context else k)
        elif isinstance(data, list):
            for i, item in enumerate(data):
                yield from find_entries(item, key, context=f'{context}[{i}]')

    for context, processing in find_entries(transform['transform'], 'transform'):
        processing_ = processing
        if not isinstance(processing, list):
            processing_ = [processing]

        for step in processing_:
            fn_name = step.get('for_each') or step.get('method')
            step.update(
                processing_functions.translate_def_to_fn(
                    step,
                    f'{context}, Processing Function: '
                    f'{processing_functions._snake_to_titlecase(fn_name)}',  # noqa: SLF001
                )
            )

    match transform['type'].lower():
        case 'indicator':
            return IndicatorTransformModel(**transform['transform'])
        case 'group':
            return GroupTransformModel(**transform['transform'])
        case _:
            ex_msg = f'Unknown transform type: {transform["type"]}'
            raise TypeError(ex_msg)


class ParamDefinition(TypedDict):
    """Parameter definition for use in the transform builder UI."""

    default: str | None
    help: str
    label: str
    name: str
    required: bool
    type: str


class FunctionDefinition(TypedDict):
    """Function definition for use in the transform builder UI."""

    name: str
    label: str
    help: str
    params: list[ParamDefinition]


def custom_function_definition(definition: FunctionDefinition):
    """Attach a custom function definition to the function."""

    def _decorator(fn):
        fn._tcex_function_definition = definition  # noqa: SLF001
        return fn

    return _decorator


class ProcessingFunctions:
    """Predefined functions to use in transforms."""

    def __init__(self, tcex) -> None:
        """."""
        self.tcex = tcex
        self.custom_fns = {}

    @custom_function_definition(
        {
            'name': 'custom',
            'label': 'Custom',
            'help': 'Describe a custom processing function.',
            'params': [
                {
                    'default': None,
                    'name': 'name',
                    'label': 'Name',
                    'type': 'str',
                    'help': 'A unique name for this custom processing function.',
                    'required': True,
                },
                {
                    'default': None,
                    'name': 'description',
                    'label': 'Description',
                    'type': 'str',
                    'help': 'Describe the custom processing function.',
                    'required': True,
                },
            ],
        }
    )
    def custom(self, value, name: str, description: str, ti_dict, transform, **kwargs):
        """Allow for custom processing to be described."""
        fn = self.custom_fns.get(name.lower())
        if not fn:
            ex_msg = f'Custom function not implemented: {description}'
            raise NotImplementedError(ex_msg)
        return fn(value, ti_dict=ti_dict, transform=transform, **kwargs)

    def static_map(self, value, mapping: dict):
        """Map values to static values.

        If there is no matching value in the mapping the original value will be returned.
        """
        if not isinstance(mapping, dict):
            mapping = json.loads(mapping)
        return mapping.get(str(value), value)

    def value_in(self, value, values: str, delimiter: str = ','):
        """Return the value if it is in the list of values, else return None."""
        if not values.startswith('"'):
            values.replace('"', '"')
            values = f'"{values}"'

        return value if value in [v.strip() for v in json.loads(values).split(delimiter)] else None

    def format_table(self, value, column_order: str):
        """Format a markdown table.

        value should be a list of objects that all have the same attributes.  The table will contain
        one row for each object in the list, and one column for each attribute of the objects.
        """
        if column_order:
            order = [c.strip() for c in column_order.split(',')]
        else:
            order = list(value[0].keys())

        table = ''
        table += f'|{"|".join(order)}|\n'
        table += f'|{"|".join(["-" for o in order])}|\n'
        for row in value:
            table += f'|{"|".join([str(row.get(o, "")) for o in order])}|\n'

        return table

    def any_to_datetime(self, value):
        """Convert any value to a datetime object."""
        return self.tcex.util.any_to_datetime(value).isoformat()

    def append(self, value, suffix: str):
        """Append a value to the input value."""
        return f'{value}{suffix}'

    def prepend(self, value, prefix: str):
        """Prepend a value to the input value."""
        return f'{prefix}{value}'

    def replace(self, value, old_value: str, new_value: str = ''):
        """Replace a value in the input value."""
        return value.replace(old_value, new_value)

    def to_uppercase(self, value):
        """Convert value to uppercase."""
        return str.upper(value)

    def to_lowercase(self, value):
        """Convert value to lowercase."""
        return str.lower(value)

    def to_titlecase(self, value):
        """Convert value to titlecase."""
        return str.title(value)

    def truncate(self, value, length: int, append_chars: str = '...'):
        """Truncate a string."""
        return self.tcex.util.truncate_string(value, length=length, append_chars=append_chars)

    def hash(self, value) -> str:
        """Hash the given value."""
        return hashlib.sha256(str(value).encode('utf-8')).hexdigest()

    def split(self, value, delimiter: str = ','):
        """Split a string into a list."""
        return [v.strip() for v in value.split(delimiter)]

    def remove_surrounding_whitespace(self, value):
        """Strip leading and trailing whitespace from a string."""
        return value.strip()

    @custom_function_definition({'name': 'uuid5', 'label': 'To UUID5', 'help': '', 'params': []})
    def uuid5(self, value, namespace=None) -> str:
        """Generate a UUID5."""
        return str(uuid.uuid5(namespace or uuid.NAMESPACE_DNS, value))

    def convert_to_MITRE_tag(self, value) -> str | None:  # noqa: N802
        """Transform MITRE tags to TC format."""
        return self.tcex.api.tc.v3.mitre_tags.get_by_id_regex(value, value)

    def translate_def_to_fn(self, api_def: dict, context: str):
        """Translate a function definition in transform builder/API format to an actual function."""
        additional_context = ''
        try:
            translated = api_def.copy()

            type_ = 'method' if 'method' in api_def else 'for_each'

            if not type_:
                ex_msg = 'No method or for_each key found in definition.'
                raise ValueError(ex_msg)  # noqa: TRY301

            fn_name = api_def[type_]

            if callable(fn_name):
                return api_def

            fn = getattr(self, fn_name)

            if not fn:
                ex_msg = f'Unknown function: {fn_name}'
                raise ValueError(ex_msg)  # noqa: TRY301

            translated[type_] = fn

            if 'kwargs' in api_def:
                sig = signature(fn)

                for kwarg in api_def['kwargs']:
                    if kwarg not in sig.parameters:
                        ex_msg = f'Unknown argument {kwarg} for function {fn_name}'
                        raise ValueError(ex_msg)  # noqa: TRY301

                    annotation = sig.parameters[kwarg].annotation
                    additional_context = f', Argument: {kwarg}'
                    match annotation():
                        case dict():
                            translated['kwargs'][kwarg] = json.loads(api_def['kwargs'][kwarg])
                        case _:
                            translated['kwargs'][kwarg] = sig.parameters[kwarg].annotation(
                                api_def['kwargs'][kwarg]
                            )
        except Exception as ex:
            # first-party
            from tcex.api.tc.ti_transform import TransformException

            ex_msg = f'{context}{additional_context}'
            raise TransformException(ex_msg, ex, context=api_def) from ex
        else:
            return translated

    def get_function_definitions(self) -> list[FunctionDefinition]:
        """Get function definitions in JSON format, suitable for the transform builder UI."""

        def _is_function(obj):
            return type(obj).__name__ == 'method'

        fns = [
            fn
            for fn in (
                getattr(self, n)
                for n in dir(self)
                if not n.startswith('_')
                and n not in ('get_function_definitions', 'translate_def_to_fn')
            )
            if _is_function(fn)
        ]

        return [
            getattr(fn, '_tcex_function_definition', {})
            or {
                'name': fn.__name__,
                'label': self._snake_to_titlecase(fn.__name__),
                'params': self._get_params_defs(fn),
                'help': getattr(fn, '__doc__', ''),
            }
            for fn in fns
        ]  # type: ignore

    @staticmethod
    def _snake_to_titlecase(name):
        return name.replace('_', ' ').title()

    @staticmethod
    def _get_params_defs(fn) -> list[ParamDefinition]:
        """Get the arguments for a function.

        Args:
            fn (function): The function to get the arguments for.

        Returns:
            list: A list of dictionaries containing the argument name and type.
        """
        sig = signature(fn)
        params = [n for n in sig.parameters if n not in ('self', 'value')]
        return [
            {
                'default': (
                    sig.parameters[p].default if sig.parameters[p].default != _empty else None
                ),
                'name': p,
                'label': ProcessingFunctions._snake_to_titlecase(p),
                'type': (
                    sig.parameters[p].annotation.__name__ if sig.parameters[p].annotation else 'str'
                ),
                'help': '',
                'required': True,
            }
            for p in params
        ]



"""TcEx Framework Module"""

# standard library
from collections.abc import Callable

from .singleton import Singleton


class Event(metaclass=Singleton):
    """Event Class"""

    def __init__(self):
        """Initialize instance properties."""
        self.channels = {}

    def send(self, channel: str, **kwargs):
        """Send message to channel."""
        for callback in self.channels.get(channel, []):
            callback(**kwargs)

    def subscribe(self, channel: str, callback: Callable):
        """Subscribe to a channel with a callback."""
        self.channels.setdefault(channel, [])
        self.channels[channel].append(callback)

    def unsubscribe(self, channel: str, callback: Callable):
        """Subscribe to a channel with a callback."""
        self.channels[channel].remove(callback)



"""TcEx Framework Module"""

# standard library
import logging
from functools import cached_property

from ..app.config.install_json import InstallJson
from ..input.model.module_requests_session_model import ModuleRequestsSessionModel
from ..pleb.proxies import proxies
from ..pleb.scoped_property import scoped_property
from ..registry import registry
from .auth.hmac_auth import HmacAuth
from .auth.tc_auth import TcAuth
from .auth.token_auth import TokenAuth
from .tc_session import TcSession

# get logger
_logger = logging.getLogger(__name__.split('.', maxsplit=1)[0])


class RequestsTc:
    """Requests Session Class"""

    def __init__(self, model: ModuleRequestsSessionModel):
        """Initialize instance properties."""
        self.model = model

        # properties
        self.install_json = InstallJson()
        self.log = _logger

    def get_session(
        self,
        auth: HmacAuth | TokenAuth | TcAuth | None = None,
        base_url: str | None = None,
        log_curl: bool | None = None,
        proxies: dict[str, str] | None = None,
        proxies_enabled: bool | None = None,
        verify: bool | str | None = None,
    ) -> TcSession:
        """Return an instance of Requests Session configured for the ThreatConnect API.

        No args are required to get a working instance of TC Session instance.

        This method allows for getting a new instance of TC Session instance. This can be
        very useful when connecting between multiple TC instances (e.g., migrating data).
        """
        if log_curl is None:
            log_curl = self.model.tc_log_curl

        if proxies_enabled is None:
            proxies_enabled = self.model.tc_proxy_tc

        if verify is None:
            verify = self.model.tc_verify

        tc_token = None
        # 1. if token module is available, use token callback
        # 2. if token is set in the model, use that (no renewal)
        # 3. no token is not available, use api credentials
        if hasattr(registry.app, 'token') and self.install_json.is_external_app is False:
            # token module is only available on tcex, not tcex-app-testing, or tcex-cli
            tc_token = registry.app.token.get_token  # type: ignore
        elif self.model.tc_token is not None:
            tc_token = self.model.tc_token

        auth = auth or TcAuth(
            tc_api_access_id=self.model.tc_api_access_id,
            tc_api_secret_key=self.model.tc_api_secret_key,
            tc_token=tc_token,
        )

        return TcSession(
            auth=auth,
            base_url=base_url or self.model.tc_api_path,
            log_curl=log_curl,
            proxies=proxies or self.proxies,
            proxies_enabled=proxies_enabled,
            user_agent=registry.app.user_agent,  # type: ignore
            verify=verify,
        )

    @cached_property
    def proxies(self) -> dict:
        """Return proxies dictionary for use with the Python Requests module."""
        return proxies(
            proxy_host=self.model.tc_proxy_host,
            proxy_port=self.model.tc_proxy_port,
            proxy_user=self.model.tc_proxy_username,
            proxy_pass=self.model.tc_proxy_password,
        )

    @scoped_property
    def session(self) -> TcSession:
        """Return an instance of Requests Session configured for the ThreatConnect API."""
        return self.get_session()



"""TcEx Framework Module"""

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class VictimAssetModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='VictimAsset Model',
    validate_assignment=True,
):
    """Victim_Asset Model"""

    _associated_type = PrivateAttr(default=True)
    _cm_type = PrivateAttr(default=False)
    _shared_type = PrivateAttr(default=False)
    _staged = PrivateAttr(default=False)

    account_name: str | None = Field(
        None,
        applies_to=['SocialNetwork', 'NetworkAccount'],
        conditional_required=['SocialNetwork', 'NetworkAccount'],
        description='The network name.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='accountName',
    )
    address: str | None = Field(
        None,
        applies_to=['EmailAddress'],
        conditional_required=['EmailAddress'],
        description='The email address associated with the E-Mail Address asset.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='address',
    )
    address_type: str | None = Field(
        None,
        applies_to=['EmailAddress'],
        description='The type of the E-Mail Address asset.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='addressType',
    )
    associated_groups: 'GroupsModel' = Field(
        None,
        description='A list of groups that this victim asset is associated with.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedGroups',
    )
    id: int | None = Field(  # type: ignore
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    network_type: str | None = Field(
        None,
        applies_to=['NetworkAccount'],
        conditional_required=['NetworkAccount'],
        description='The type of network.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='networkType',
    )
    phone: str | None = Field(
        None,
        applies_to=['Phone'],
        conditional_required=['Phone'],
        description='The phone number of the asset.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='phone',
    )
    social_network: str | None = Field(
        None,
        applies_to=['SocialNetwork'],
        conditional_required=['SocialNetwork'],
        description='The type of social network.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='socialNetwork',
    )
    type: str | None = Field(
        None,
        description='Type of victim asset.',
        methods=['POST'],
        read_only=False,
        title='type',
    )
    victim_id: int | None = Field(
        None,
        description='Victim id of victim asset.',
        methods=['POST'],
        read_only=False,
        title='victimId',
    )
    web_link: str | None = Field(
        None,
        allow_mutation=False,
        description='A link to the ThreatConnect details page for this entity.',
        read_only=True,
        title='webLink',
    )
    website: str | None = Field(
        None,
        applies_to=['WebSite'],
        conditional_required=['WebSite'],
        description='The website of the asset.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='website',
    )

    @validator('associated_groups', always=True, pre=True)
    @classmethod
    def _validate_groups(cls, v):
        if not v:
            return GroupsModel()  # type: ignore
        return v


class VictimAssetDataModel(
    BaseModel,
    title='VictimAsset Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Victim_Assets Data Model"""

    data: list[VictimAssetModel] | None = Field(
        [],
        description='The data for the VictimAssets.',
        methods=['POST', 'PUT'],
        title='data',
    )


class VictimAssetsModel(
    BaseModel,
    title='VictimAssets Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Victim_Assets Model"""

    _mode_support = PrivateAttr(default=False)

    data: list[VictimAssetModel] | None = Field(
        [],
        description='The data for the VictimAssets.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# first-party
from tcex.api.tc.v3.groups.group_model import GroupsModel

# add forward references
VictimAssetDataModel.update_forward_refs()
VictimAssetModel.update_forward_refs()
VictimAssetsModel.update_forward_refs()



"""TcEx Framework Module"""

# standard library

# third-party
from pydantic import Field, validator

# first-party
from tcex.api.tc.v3.security.assignee_user_group_model import AssigneeUserGroupModel
from tcex.api.tc.v3.security.assignee_user_model import AssigneeUserModel
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class AssigneeModel(
    V3ModelABC,
    title='User Data Model',
    alias_generator=Util().snake_to_camel,
    fields={'id': {'exclude': True}},
    validate_assignment=True,
):
    """Assignee Model"""

    type: str | None = Field(
        None,
        description='The **Type** for the Assignee.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='type',
    )

    data: AssigneeUserModel | AssigneeUserGroupModel | None = Field(
        None,
        description='The **Data** for the Assignee.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='data',
    )

    @validator('type', always=True)
    @classmethod
    def _validate_type(cls, v):
        if not v:
            return v

        if v.lower() == 'user':
            return 'User'

        if v.lower() in ['group', 'user_group', 'usergroup']:
            return 'Group'

        ex_msg = 'Value must be either `User` or `Group`.'
        raise ValueError(ex_msg)



"""Custom functions for ThreatConnect Jmespath Playbook App."""

# standard library
import collections
import itertools
import uuid
from copy import deepcopy
from typing import Any

# third-party
import jmespath
from jmespath import exceptions, functions
from jmespath.visitor import _Expression

from ..util import Util


def jmespath_options() -> jmespath.Options:
    """Return the jmespath options."""
    return jmespath.Options(custom_functions=TcFunctions(), dict_cls=collections.OrderedDict)


class TcFunctions(functions.Functions):
    """ThreatConnect custom jmespath functions."""

    def _get_expression_key(self, expref: _Expression):
        """Return the key from the expression."""
        match expref.expression['type']:  # type: ignore
            case 'field':
                return expref.expression['value']  # type: ignore

            case 'subexpression':
                return expref.expression.get('children')[-1]['value']  # type: ignore

            case _:
                ex_msg = f"""Invalid expression type of {expref.expression['type']}."""  # type: ignore

                raise RuntimeError(ex_msg)

    def _update_expression_parent(self, expref: _Expression, key: str, item: dict, value: Any):
        """Return the key from the expression."""
        match expref.expression['type']:  # type: ignore
            case 'field':
                item[key] = value

            case 'subexpression':
                expand_item = item
                children = expref.expression.get('children') or []  # type: ignore

                for index, child in enumerate(children):
                    child_value = child['value']
                    expand_item = expand_item[child_value]

                    if index == len(children) - 2:
                        expand_item[key] = value

    @functions.signature({'types': ['string']}, {'types': ['string']})
    def _func_format_datetime(self, input_: str, date_format: str):
        """Return array after popping value at address out.

        Expression:
        {
            group: [].{
                name: 'post_title',
                type: 'Event',
                eventDate: format_datetime(discovered, '%Y-%m-%dT%H:%M:%SZ')
            }
        }
        """
        return Util.any_to_datetime(input_).strftime(date_format)

    @functions.signature({'types': ['array']}, {'types': ['expref']})
    def _func_expand(self, array: list[dict], expref: _Expression):
        """Expand results into an array of objects.

        Expression:
        expand(@, &urls)

        Data:
        [
            {
                "id": "123",
                "urls": [
                    "abc.example.com",
                    "def.example.com"
                ],
                "category": "Malicious"
            }
        ]

        Output:
        [
            {
                "id": "123",
                "urls": "abc.example.com",
                "category": "Malicious"
            },
            {
                "id": "123",
                "urls": "def.example.com",
                "category": "Malicious"
            }
        ]
        """
        # short circuit, if no array
        if not array:
            return array

        key_func = self._create_key_func(expref, ['array'], 'expand2')  # type: ignore

        expression_key = self._get_expression_key(expref)

        result = []
        for item in array:
            for key in key_func(item):
                new_item = deepcopy(item)
                self._update_expression_parent(expref, expression_key, new_item, key)
                result.append(new_item)
        return result

    @functions.signature({'types': ['array']}, {'types': ['expref']})
    def _func_group_by(self, array: list[dict], expref: _Expression):
        """Group results into an array of objects.

        Expression:
        group_by(items, &spec.nodeName)

        Data:
        {
          "items": [
            {
              "spec": {
                "nodeName": "node_01",
                "other": "values_01"
              }
            },
            {
              "spec": {
                "nodeName": "node_02",
                "other": "values_02"
              }
            },
            {
              "spec": {
                "nodeName": "node_01",
                "other": "values_04"
              }
            }
          ]
        }

        Output:
        {
            "node_01": [
                {
                    "spec": {
                        "nodeName": "node_01",
                        "other": "values_01"
                    }
                },
                {
                    "spec": {
                        "nodeName": "node_01",
                        "other": "values_04"
                    }
                }
            ],
            "node_02": [
                {
                    "spec": {
                        "nodeName": "node_02",
                        "other": "values_02"
                    }
                }
            ]
        }
        """
        # short circuit, if no array
        if not array:
            return array

        key_func = self._create_key_func(expref, ['null', 'string'], 'group_by')  # type: ignore

        result = {}
        for item in array:
            result.setdefault(key_func(item), []).append(item)
        return result

    @functions.signature({'types': ['array']}, {'types': ['string']})
    def _func_delete(self, arr: list, search: str):
        """Return array of objects after removing key for each object.

        Expression:
        delete([], 'state')
        """
        for a in arr:
            a.pop(search, None)
        return arr

    @functions.signature({'types': ['array']}, {'types': ['string']})
    def _func_null_leaf(self, arr: list, search: str):
        """Return value in array even if they are null.

        Expression:
        locations[] | null_leaf(@, 'state')
        """
        return [a.get(search) for a in arr]

    @functions.signature({'types': ['string']}, {'types': ['string']}, {'types': ['string']})
    def _func_replace(self, input_: str, search: str, replace: str, count: int = -1):
        """Replace occurrences of search with replace in the input string.

        Expression:
        replace('hello world', 'world', 'there')

        Output:
        'hello there'
        """
        return input_.replace(search, replace, count)

    @functions.signature({'types': ['string']})
    def _func_uuid5(self, input_: str):
        """Return array after popping value at address out.

        Expression:
        {
            group: [].{
                name: 'post_title',
                type: 'Event',
                eventDate: format_datetime(discovered, '%Y-%m-%dT%H:%M:%SZ')
                xid: uuid5(post_title)
            }
        }
        """
        return str(uuid.uuid5(uuid.NAMESPACE_DNS, input_))

    @functions.signature(
        {'types': ['array']},
        {'types': ['null', 'string'], 'optional': True},  # type: ignore
    )
    def _func_zip(self, arrays: list[list], fill_value: str | None = None):
        """Return array after popping value at address out.

        Expression:
        {data: zip([first_names, last_names, ages])}

        Data:
        {
            "first_names": ["bob", "joe", "sally"],
            "last_names": ["smith", "jones", "blah"]
            "ages": ["32", "41"]
        }

        Output
        {
            data: [
                [bob, smith, 32],
                [joe, jones, 41],
                [sally, blah, null]
            ]
        }
        """
        return list(itertools.zip_longest(*arrays, fillvalue=fill_value))

    @functions.signature(
        {'types': ['array']},
        {'types': ['array']},
        {'types': ['null', 'string'], 'optional': True},  # type: ignore
    )
    def _func_zip_to_objects(
        self,
        keys: list[str],
        values: list[list],
        fill_value: str | None = None,
    ):
        """Return array after popping value at address out.

        Expression:
        {data: zip_to_object(['name', 'last_name', 'age'], [first_names, last_names, ages]}

        Data:
        {
            "first_names": ["bob", "joe", "sally"],
            "last_names": ["smith", "jones", "blah"]
            "ages": ["30", "40"]
        }

        Output:
        {
            data: [
                {
                    first_name: bob,
                    last_name: smith,
                    age: 30
                },
                {
                    first_name: joe,
                    last_name: jones,
                    age: 40
                },
                {
                    first_name: sally,
                    last_name: blah,
                    age: null
                }
            ]
        }
        """
        if len(keys) != len(values):
            ex_msg = 'Keys and values must be the same length.'
            raise RuntimeError(ex_msg)

        data = []
        values = itertools.zip_longest(*values, fillvalue=fill_value)  # type: ignore
        for value in values:
            data.append({k: value[i] for i, k in enumerate(keys)})
        return data

    #
    # Update jmespath Functions plumbing to support optional inputs
    #

    def _type_check(self, actual: list, signature: tuple, function_name: str):
        """Check the arg type to signature type."""
        for i in range(min(len(signature), len(actual))):
            allowed_types = signature[i]['types']
            if allowed_types:
                self._type_check_single(actual[i], allowed_types, function_name)  # type: ignore

    def _validate_arguments(self, args: list, signature: tuple, function_name: str):
        """Check the provided args match the signature type, taking into account optional args."""
        # short circuit, if no signature
        if len(signature) == 0:
            return self._type_check(args, signature, function_name)

        # get the number of required and optional arguments
        required_arguments_count = len(
            [param for param in signature if param.get('optional', False) is not True]
        )
        optional_arguments_count = len(
            [param for param in signature if param.get('optional', False) is not True]
        )
        has_variadic = signature[-1].get('variadic', False)

        if has_variadic:
            if len(args) < len(signature):
                raise exceptions.VariadictArityError(len(signature), len(args), function_name)
        elif optional_arguments_count > 0:
            if len(args) < required_arguments_count or len(args) > (
                required_arguments_count + optional_arguments_count
            ):
                raise exceptions.ArityError(len(signature), len(args), function_name)
        elif len(args) != required_arguments_count:
            raise exceptions.ArityError(len(signature), len(args), function_name)
        return self._type_check(args, signature, function_name)



"""TcEx Framework Module"""

# standard library
from enum import Enum

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql_type import TqlType


class AttributeTypeFilter(FilterABC):
    """Filter Object for AttributeTypes"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.ATTRIBUTE_TYPES.value

    def associated_type(self, operator: Enum, associated_type: list | str):
        """Filter Associated Type based on **associatedType** keyword.

        Args:
            operator: The operator enum for the filter.
            associated_type: The data type(s) that the attribute type can be used for.
        """
        if isinstance(associated_type, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('associatedType', operator, associated_type, TqlType.STRING)

    def default(self, operator: Enum, default: bool):
        """Filter Displayed based on **default** keyword.

        Args:
            operator: The operator enum for the filter.
            default: Whether or not the attribute type is displayable on the item.
        """
        self._tql.add_filter('default', operator, default, TqlType.BOOLEAN)

    def default_owner_id(self, operator: Enum, default_owner_id: int | list):
        """Filter Attribute settings owner id based on **defaultOwnerId** keyword.

        Args:
            operator: The operator enum for the filter.
            default_owner_id: The owner id of the attribute type settings.
        """
        if isinstance(default_owner_id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('defaultOwnerId', operator, default_owner_id, TqlType.INTEGER)

    def default_type(self, operator: Enum, default_type: list | str):
        """Filter Defaulted Type based on **defaultType** keyword.

        Args:
            operator: The operator enum for the filter.
            default_type: The data type(s) that the attribute type is defaulted for.
        """
        if isinstance(default_type, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('defaultType', operator, default_type, TqlType.STRING)

    def description(self, operator: Enum, description: list | str):
        """Filter Description based on **description** keyword.

        Args:
            operator: The operator enum for the filter.
            description: The description of the attribute type.
        """
        if isinstance(description, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('description', operator, description, TqlType.STRING)

    def id(self, operator: Enum, id: int | list):  # noqa: A002
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the attribute type.
        """
        if isinstance(id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def maxsize(self, operator: Enum, maxsize: int | list):
        """Filter Maxsize based on **maxsize** keyword.

        Args:
            operator: The operator enum for the filter.
            maxsize: Max size of the attribute.
        """
        if isinstance(maxsize, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('maxsize', operator, maxsize, TqlType.INTEGER)

    def name(self, operator: Enum, name: list | str):
        """Filter Name based on **name** keyword.

        Args:
            operator: The operator enum for the filter.
            name: The name of the attribute type.
        """
        if isinstance(name, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('name', operator, name, TqlType.STRING)

    def owner(self, operator: Enum, owner: int | list):
        """Filter Owner ID based on **owner** keyword.

        Args:
            operator: The operator enum for the filter.
            owner: The owner ID of the attribute type.
        """
        if isinstance(owner, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('owner', operator, owner, TqlType.INTEGER)

    def owner_name(self, operator: Enum, owner_name: list | str):
        """Filter Owner Name based on **ownerName** keyword.

        Args:
            operator: The operator enum for the filter.
            owner_name: The owner name of the attribute type.
        """
        if isinstance(owner_name, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('ownerName', operator, owner_name, TqlType.STRING)

    def system(self, operator: Enum, system: bool):
        """Filter SystemLevel based on **system** keyword.

        Args:
            operator: The operator enum for the filter.
            system: A flag to show System level attributes (TRUE) or owner-specific ones only
                (FALSE).
        """
        self._tql.add_filter('system', operator, system, TqlType.BOOLEAN)



"""TcEx Framework Module"""

# standard library
from typing import TYPE_CHECKING, Any, cast

# third-party
import wrapt

# first-party
from tcex.tcex import TcEx

if TYPE_CHECKING:
    # standard library
    from collections.abc import Callable


class Debug:
    """Debug function inputs.

    This decorator will log the inputs to the function to assist in debugging an App.

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        @Debug()
        def my_method(arg1, args2):
            print(arg1, arg2)
    """

    @wrapt.decorator
    def __call__(self, *wrapped_args) -> Any:  # noqa: D417
        """Implement __call__ function for decorator.

        Args:
            wrapped: The wrapped function which in turns
                needs to be called by your wrapper function.
            instance: The object to which the wrapped
                function was bound when it was called.
            args: The list of positional arguments supplied
                when the decorated function was called.
            kwargs: The dictionary of keyword arguments
                supplied when the decorated function was called.

        Returns:
            function: The custom decorator function.
        """
        # using wrapped args to support typing hints in PyRight
        wrapped: Callable = wrapped_args[0]
        app: Any = wrapped_args[1]
        args: list = wrapped_args[2] if len(wrapped_args) > 1 else []
        kwargs: dict = wrapped_args[3] if len(wrapped_args) > 2 else {}  # noqa: PLR2004

        def debug() -> Any:
            """Iterate over data, calling the decorated function for each value.

            Args:
                app (class): The instance of the App class "self".
                *args: Additional positional arguments.
                **kwargs: Additional keyword arguments.
            """
            tcex = cast(TcEx, app.tcex)

            data = wrapped(*args, **kwargs)
            tcex.log.debug(
                f'function: "{self.__class__.__name__}", args: "{args}", kwargs: "{kwargs}"'
            )
            return data

        return debug()



"""TcEx Framework Module"""

from .app_spec_yml_model import AppSpecYmlModel
from .install_json_model import InstallJsonModel
from .job_json_model import JobJsonModel
from .layout_json_model import LayoutJsonModel
from .tcex_json_model import TcexJsonModel

__all__ = [
    'AppSpecYmlModel',
    'InstallJsonModel',
    'JobJsonModel',
    'LayoutJsonModel',
    'TcexJsonModel',
]



"""TcEx Framework Module"""

# third-party
from pydantic import Field, PrivateAttr

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class KeywordSectionModel(
    V3ModelABC,
    title='Keyword Section Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Model Definition

    {
        "compareValue": "includes",
        "keywords": [
            {
                "value": "test"
            }
        ]
    }
    """

    _associated_type = PrivateAttr(default=False)
    _cm_type = PrivateAttr(default=False)
    _shared_type = PrivateAttr(default=False)
    _staged = PrivateAttr(default=False)

    section_number: int | None = Field(
        None,
        description='The section number of the keyword section.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='sectionNumber',
    )
    compare_value: str | None = Field(
        None,
        description='The compare value for the keyword section.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='compareValue',
    )
    keywords: list[dict] | None = Field(
        [],
        description='A list of keywords for the keyword section.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='keywords',
    )



"""TcEx Framework Module"""

from .singleton import Singleton


class NoneModel(metaclass=Singleton):
    """A dummy model that return None for all attribute requests."""

    def __getattribute__(self, _: str):
        """Return None for any attribute request."""
        return



"""TcEx Framework Module"""

# standard library
from collections.abc import Generator, Iterator
from typing import TYPE_CHECKING

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.groups.group_model import GroupModel
from tcex.api.tc.v3.indicators.indicator_model import IndicatorModel
from tcex.api.tc.v3.intel_requirements.intel_requirement_filter import IntelRequirementFilter
from tcex.api.tc.v3.intel_requirements.intel_requirement_model import (
    IntelRequirementModel,
    IntelRequirementsModel,
)
from tcex.api.tc.v3.intel_requirements.keyword_sections.keyword_section_model import (
    KeywordSectionModel,
)
from tcex.api.tc.v3.object_abc import ObjectABC
from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC
from tcex.api.tc.v3.tags.tag_model import TagModel
from tcex.api.tc.v3.victim_assets.victim_asset_model import VictimAssetModel

if TYPE_CHECKING:  # pragma: no cover
    # first-party
    from tcex.api.tc.v3.artifacts.artifact import Artifact  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.cases.case import Case  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.groups.group import Group  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.indicators.indicator import Indicator  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.tags.tag import Tag  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.victim_assets.victim_asset import VictimAsset  # CIRCULAR-IMPORT


class IntelRequirement(ObjectABC):
    """IntelRequirements Object.

    Args:
        associated_artifacts (Artifacts, kwargs): A list of Artifacts associated with this Group.
        associated_cases (Cases, kwargs): A list of Cases associated with this Group.
        associated_groups (Groups, kwargs): A list of groups associated with this group.
        associated_indicators (Indicators, kwargs): A list of indicators associated with this group.
        associated_victim_assets (VictimAssets, kwargs): A list of victim assets associated with
            this group.
        category (IntelReqType, kwargs): The category of the intel requirement.
        description (str, kwargs): The description of the intel requirement.
        keyword_sections (array, kwargs): The section of the intel requirement that contains the
            keywords.
        requirement_text (str, kwargs): The detailed text of the intel requirement.
        reset_results (bool, kwargs): Flag to reset results when updating keywords.
        subtype (IntelReqType, kwargs): The subtype of the intel requirement.
        tags (Tags, kwargs): A list of Tags corresponding to the item (NOTE: Setting this parameter
            will replace any existing tag(s) with the one(s) specified).
        unique_id (str, kwargs): The unique id of the intel requirement.
        xid (str, kwargs): The xid of the item.
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(kwargs.pop('session', None))

        # properties
        self._model: IntelRequirementModel = IntelRequirementModel(**kwargs)
        self._nested_field_name = 'intelRequirements'
        self._nested_filter = 'has_intel_requirement'
        self.type_ = 'Intel Requirement'

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.INTEL_REQUIREMENTS.value

    @property
    def model(self) -> IntelRequirementModel:
        """Return the model data."""
        return self._model

    @model.setter
    def model(self, data: dict | IntelRequirementModel):
        """Create model using the provided data."""
        if isinstance(data, type(self.model)):
            # provided data is already a model, nothing required to change
            self._model = data
        elif isinstance(data, dict):
            # provided data is raw response, load the model
            self._model = type(self.model)(**data)
        else:
            ex_msg = f'Invalid data type: {type(data)} provided.'
            raise RuntimeError(ex_msg)  # noqa: TRY004

    @property
    def as_entity(self) -> dict:
        """Return the entity representation of the object."""
        type_ = self.type_

        return {'type': type_, 'id': self.model.id, 'value': self.model.requirement_text}

    @property
    def associated_artifacts(self) -> Generator['Artifact', None, None]:
        """Yield Artifact from Artifacts."""
        # first-party
        from tcex.api.tc.v3.artifacts.artifact import Artifacts

        yield from self._iterate_over_sublist(Artifacts)  # type: ignore

    @property
    def associated_cases(self) -> Generator['Case', None, None]:
        """Yield Case from Cases."""
        # first-party
        from tcex.api.tc.v3.cases.case import Cases

        yield from self._iterate_over_sublist(Cases)  # type: ignore

    @property
    def associated_groups(self) -> Generator['Group', None, None]:
        """Yield Group from Groups."""
        # first-party
        from tcex.api.tc.v3.groups.group import Groups

        yield from self._iterate_over_sublist(Groups)  # type: ignore

    @property
    def associated_indicators(self) -> Generator['Indicator', None, None]:
        """Yield Indicator from Indicators."""
        # first-party
        from tcex.api.tc.v3.indicators.indicator import Indicators

        yield from self._iterate_over_sublist(Indicators)  # type: ignore

    @property
    def associated_victim_assets(self) -> Generator['VictimAsset', None, None]:
        """Yield VictimAsset from VictimAssets."""
        # first-party
        from tcex.api.tc.v3.victim_assets.victim_asset import VictimAssets

        yield from self._iterate_over_sublist(VictimAssets)  # type: ignore

    @property
    def tags(self) -> Generator['Tag', None, None]:
        """Yield Tag from Tags."""
        # first-party
        from tcex.api.tc.v3.tags.tag import Tags

        yield from self._iterate_over_sublist(Tags)  # type: ignore

    def stage_associated_group(self, data: dict | ObjectABC | GroupModel):
        """Stage group on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = GroupModel(**data)

        if not isinstance(data, GroupModel):
            ex_msg = 'Invalid type passed in to stage_associated_group'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.associated_groups.data.append(data)  # type: ignore

    def stage_associated_victim_asset(self, data: dict | ObjectABC | VictimAssetModel):
        """Stage victim_asset on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = VictimAssetModel(**data)

        if not isinstance(data, VictimAssetModel):
            ex_msg = 'Invalid type passed in to stage_associated_victim_asset'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.associated_victim_assets.data.append(data)  # type: ignore

    def stage_associated_indicator(self, data: dict | ObjectABC | IndicatorModel):
        """Stage indicator on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = IndicatorModel(**data)

        if not isinstance(data, IndicatorModel):
            ex_msg = 'Invalid type passed in to stage_associated_indicator'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.associated_indicators.data.append(data)  # type: ignore

    def replace_keyword_section(self, data: dict | list | ObjectABC | KeywordSectionModel):
        """Replace keyword_section on the object."""
        if not isinstance(data, list):
            data = [data]

        if all(isinstance(item, (KeywordSectionModel | ObjectABC)) for item in data):
            transformed_data = data
        elif all(isinstance(item, dict) for item in data):
            transformed_data = [KeywordSectionModel(**d) for d in data]
        else:
            ex_msg = 'Invalid data to replace_keyword_section'
            raise ValueError(ex_msg)

        for item in transformed_data:
            item._staged = True  # noqa: SLF001

        self.model.keyword_sections = transformed_data  # type: ignore

    def stage_tag(self, data: dict | ObjectABC | TagModel):
        """Stage tag on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = TagModel(**data)

        if not isinstance(data, TagModel):
            ex_msg = 'Invalid type passed in to stage_tag'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.tags.data.append(data)  # type: ignore


class IntelRequirements(ObjectCollectionABC):
    """IntelRequirements Collection.

    # Example of params input
    {
        "result_limit": 100,  # Limit the retrieved results.
        "result_start": 10,  # Starting count used for pagination.
        "fields": ["caseId", "summary"]  # Select additional return fields.
    }

    Args:
        session (Session): Session object configured with TC API Auth.
        tql_filters (list): List of TQL filters.
        params (dict): Additional query params (see example above).
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(
            kwargs.pop('session', None), kwargs.pop('tql_filter', None), kwargs.pop('params', None)
        )
        self._model = IntelRequirementsModel(**kwargs)
        self.type_ = 'intel_requirements'

    def __iter__(self) -> Iterator[IntelRequirement]:
        """Return CM objects."""
        return self.iterate(base_class=IntelRequirement)  # type: ignore

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.INTEL_REQUIREMENTS.value

    @property
    def filter(self) -> IntelRequirementFilter:
        """Return the type specific filter object."""
        return IntelRequirementFilter(self.tql)



"""TcEx Framework Module"""

# standard library
import gzip
import os
import shutil
from logging.handlers import RotatingFileHandler
from pathlib import Path


class RotatingFileHandlerCustom(RotatingFileHandler):
    """Logger handler for ThreatConnect Exchange File logging."""

    def __init__(
        self,
        filename: str,
        mode: str = 'a',
        maxBytes: int = 0,  # noqa: N803
        backupCount: int = 0,  # noqa: N803
        encoding: str | None = None,
        delay: bool = False,
    ):
        """Customize RotatingFileHandler to create full log path.

        Args:
            filename: The name of the logfile.
            mode: The write mode for the file.
            maxBytes: The max file size before rotating.
            backupCount: The maximum # of backup files.
            encoding: The log file encoding.
            delay: If True, then file opening is deferred until the first call to emit().
        """
        if encoding is None and os.getenv('LANG') is None:
            encoding = 'UTF-8'

        Path(filename).parent.mkdir(parents=True, exist_ok=True)
        RotatingFileHandler.__init__(self, filename, mode, maxBytes, backupCount, encoding, delay)

        # set namer
        self.namer = self.custom_gzip_namer
        self.rotator = self.custom_gzip_rotator

    @staticmethod
    def custom_gzip_namer(name):
        """Namer for rotating log handler with gz extension.

        Args:
            name: The current name of the logfile.
        """
        return name + '.gz'

    @staticmethod
    def custom_gzip_rotator(source: str, dest: str):
        """Rotate and compress log file.

        Args:
            source: The source filename.
            dest: The destination filename.
        """
        source_filename = Path(source)
        with source_filename.open(mode='rb') as f_in, gzip.open(dest, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
        source_filename.unlink()



"""TcEx Framework Module"""

# standard library
import json
from collections.abc import Callable


class Tag:
    """ThreatConnect Batch Tag Object"""

    __slots__ = ['_tag_data', '_valid']

    def __init__(self, name: str, formatter: Callable[[str], str] | None = None):
        """Initialize instance properties.

        Args:
            name: The value for this tag.
            formatter: A callable that take a tag value and returns a formatted tag.
        """
        if formatter is not None:
            name = formatter(name)
        self._tag_data = {'name': name}
        # is tag not null or ''
        self._valid = True
        if not name:
            self._valid = False

    @property
    def data(self) -> dict:
        """Return Tag data."""
        return self._tag_data

    @property
    def name(self) -> str:
        """Return Tag name."""
        return self._tag_data.get('name')  # type: ignore

    @property
    def valid(self) -> bool:
        """Return valid data."""
        return self._valid

    def __str__(self) -> str:
        """Return string representation of object."""
        return json.dumps(self.data, indent=4)



"""TcEx Framework Module"""

# standard library
from enum import Enum, unique


@unique
class TqlType(Enum):
    """ThreatConnect API TQL Types"""

    BOOLEAN = 'Boolean'
    FLOAT = 'Float'
    INTEGER = 'Integer'
    STRING = 'String'
    SUB_QUERY = 'Sub Query'



"""TcEx Framework Module"""

# standard library
from datetime import datetime

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class IndicatorModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='Indicator Model',
    validate_assignment=True,
):
    """Indicator Model"""

    _associated_type = PrivateAttr(default=True)
    _cm_type = PrivateAttr(default=False)
    _shared_type = PrivateAttr(default=False)
    _staged = PrivateAttr(default=False)

    active: bool = Field(
        None,
        description='Is the indicator active?',
        methods=['POST', 'PUT'],
        read_only=False,
        title='active',
    )
    active_locked: bool = Field(
        None,
        description='Lock the indicator active value?',
        methods=['POST', 'PUT'],
        read_only=False,
        title='activeLocked',
    )
    address: str | None = Field(
        None,
        applies_to=['EmailAddress'],
        conditional_required=['EmailAddress'],
        description=(
            'The email address associated with this indicator (EmailAddress specific summary '
            'field).'
        ),
        methods=['POST'],
        read_only=False,
        title='address',
    )
    associated_artifacts: 'ArtifactsModel' = Field(
        None,
        description='A list of Artifacts associated with this Indicator.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedArtifacts',
    )
    associated_cases: 'CasesModel' = Field(
        None,
        description='A list of Cases associated with this Indicator.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedCases',
    )
    associated_groups: 'GroupsModel' = Field(
        None,
        description='A list of groups that this indicator is associated with.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedGroups',
    )
    associated_indicators: 'IndicatorsModel' = Field(
        None,
        description='A list of indicators associated with this indicator.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedIndicators',
    )
    attributes: 'IndicatorAttributesModel' = Field(
        None,
        description='A list of Attributes corresponding to the Indicator.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='attributes',
    )
    confidence: int | None = Field(
        None,
        description='The indicator threat confidence.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='confidence',
    )
    custom_association_names: list[str] = Field(
        None,
        description='The custom association names assigned to this indicator.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='customAssociationNames',
    )
    custom_associations: 'IndicatorsModel' = Field(
        None,
        description='A list of indicators with custom associations to this indicator.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='customAssociations',
    )
    date_added: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The date and time that the item was first created.',
        read_only=True,
        title='dateAdded',
    )
    description: str | None = Field(
        None,
        allow_mutation=False,
        description='The indicator description text.',
        read_only=True,
        title='description',
    )
    dns_active: bool = Field(
        None,
        applies_to=['Host'],
        description='Is dns active for the indicator?',
        methods=['POST', 'PUT'],
        read_only=False,
        title='dnsActive',
    )
    dns_resolution: dict | None = Field(
        None,
        allow_mutation=False,
        applies_to=['Host', 'Address'],
        conditional_required=['Host', 'Address'],
        description='Dns resolution data for the Host or Address indicator.',
        read_only=True,
        title='dnsResolution',
    )
    enrichment: dict | None = Field(
        None,
        allow_mutation=False,
        description='Enrichment data.',
        read_only=True,
        title='enrichment',
    )
    external_date_added: datetime | None = Field(
        None,
        description='The date and time that the item was first created externally.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='externalDateAdded',
    )
    external_date_expires: datetime | None = Field(
        None,
        description='The date and time the item expires externally.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='externalDateExpires',
    )
    external_last_modified: datetime | None = Field(
        None,
        description='The date and time the item was modified externally.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='externalLastModified',
    )
    false_positive_flag: bool = Field(
        None,
        description='Is the indicator a false positive?',
        methods=['POST', 'PUT'],
        read_only=False,
        title='falsePositiveFlag',
    )
    false_positive_reported_by_user: bool = Field(
        None,
        allow_mutation=False,
        description='Has a false positive been reported by this user for this indicator today?',
        read_only=True,
        title='falsePositiveReportedByUser',
    )
    false_positives: int | None = Field(
        None,
        allow_mutation=False,
        description='The number of false positives reported for this indicator.',
        read_only=True,
        title='falsePositives',
    )
    file_actions: 'FileActionsModel' = Field(
        None,
        description='The type of file action associated with this indicator.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='fileActions',
    )
    file_occurrences: 'FileOccurrencesModel' = Field(
        None,
        description='A list of file occurrences associated with this indicator.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='fileOccurrences',
    )
    first_seen: datetime | None = Field(
        None,
        description='The date and time that the item was first seen.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='firstSeen',
    )
    geo_location: dict | None = Field(
        None,
        allow_mutation=False,
        applies_to=['Host', 'Address'],
        conditional_required=['Host', 'Address'],
        description='Geographical localization of the Host or Address indicator.',
        read_only=True,
        title='geoLocation',
    )
    host_name: str | None = Field(
        None,
        applies_to=['Host'],
        conditional_required=['Host'],
        description='The host name of the indicator (Host specific summary field).',
        methods=['POST'],
        read_only=False,
        title='hostName',
    )
    id: int | None = Field(  # type: ignore
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    investigation_links: dict | None = Field(
        None,
        allow_mutation=False,
        description=(
            'Resource links that provide additional information to assist in investigation.'
        ),
        read_only=True,
        title='investigationLinks',
    )
    ip: str | None = Field(
        None,
        applies_to=['Address'],
        conditional_required=['Address'],
        description=(
            'The ip address associated with this indicator (Address specific summary field).'
        ),
        methods=['POST'],
        read_only=False,
        title='ip',
    )
    last_false_positive: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The date and time of the last false positive reported for this indicator.',
        read_only=True,
        title='lastFalsePositive',
    )
    last_modified: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The date and time that the indicator was last modified.',
        read_only=True,
        title='lastModified',
    )
    last_observed: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The date and time that the indicator was last observed.',
        read_only=True,
        title='lastObserved',
    )
    last_seen: datetime | None = Field(
        None,
        description='The date and time that the item was last seen.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='lastSeen',
    )
    legacy_link: str | None = Field(
        None,
        allow_mutation=False,
        description='A link to the legacy ThreatConnect details page for this entity.',
        read_only=True,
        title='legacyLink',
    )
    md5: str | None = Field(
        None,
        applies_to=['File'],
        description='The md5 associated with this indicator (File specific summary field).',
        methods=['POST', 'PUT'],
        read_only=False,
        title='md5',
    )
    mode: str | None = Field(
        None,
        applies_to=['File'],
        description='The operation to perform on the file hashes (delete | merge).',
        methods=['POST', 'PUT'],
        read_only=False,
        title='mode',
    )
    observations: int | None = Field(
        None,
        description='The number of times this indicator has been observed.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='observations',
    )
    owner_id: int | None = Field(
        None,
        description='The id of the Organization, Community, or Source that the item belongs to.',
        methods=['POST'],
        read_only=False,
        title='ownerId',
    )
    owner_name: str | None = Field(
        None,
        conditional_read_only=['Victim'],
        description='The name of the Organization, Community, or Source that the item belongs to.',
        methods=['POST'],
        read_only=False,
        title='ownerName',
    )
    private_flag: bool = Field(
        None,
        description='Is this indicator private?',
        methods=['POST', 'PUT'],
        read_only=False,
        title='privateFlag',
    )
    rating: int | None = Field(
        None,
        description='The indicator threat rating.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='rating',
    )
    security_labels: 'SecurityLabelsModel' = Field(
        None,
        description=(
            'A list of Security Labels corresponding to the Intel item (NOTE: Setting this '
            'parameter will replace any existing tag(s) with the one(s) specified).'
        ),
        methods=['POST', 'PUT'],
        read_only=False,
        title='securityLabels',
    )
    sha1: str | None = Field(
        None,
        applies_to=['File'],
        description='The sha1 associated with this indicator (File specific summary field).',
        methods=['POST', 'PUT'],
        read_only=False,
        title='sha1',
    )
    sha256: str | None = Field(
        None,
        applies_to=['File'],
        description='The sha256 associated with this indicator (File specific summary field).',
        methods=['POST', 'PUT'],
        read_only=False,
        title='sha256',
    )
    size: int | None = Field(
        None,
        applies_to=['File'],
        description='The size of the file.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='size',
    )
    source: str | None = Field(
        None,
        allow_mutation=False,
        description='The source for this indicator.',
        read_only=True,
        title='source',
    )
    summary: str | None = Field(
        None,
        allow_mutation=False,
        description='The indicator summary.',
        read_only=True,
        title='summary',
    )
    tags: 'TagsModel' = Field(
        None,
        description=(
            'A list of Tags corresponding to the item (NOTE: Setting this parameter will replace '
            'any existing tag(s) with the one(s) specified).'
        ),
        methods=['POST', 'PUT'],
        read_only=False,
        title='tags',
    )
    text: str | None = Field(
        None,
        applies_to=['URL'],
        conditional_required=['URL'],
        description='The url text value of the indicator (Url specific summary field).',
        methods=['POST'],
        read_only=False,
        title='text',
    )
    threat_assess_confidence: float | None = Field(
        None,
        allow_mutation=False,
        description='The Threat Assess confidence for this indicator.',
        read_only=True,
        title='threatAssessConfidence',
    )
    threat_assess_rating: float | None = Field(
        None,
        allow_mutation=False,
        description='The Threat Assess rating for this indicator.',
        read_only=True,
        title='threatAssessRating',
    )
    threat_assess_score: int | None = Field(
        None,
        allow_mutation=False,
        description='The Threat Assess score for this indicator.',
        read_only=True,
        title='threatAssessScore',
    )
    threat_assess_score_false_positive: int | None = Field(
        None,
        allow_mutation=False,
        description='The Threat Assess score for false positives related to this indicator.',
        read_only=True,
        title='threatAssessScoreFalsePositive',
    )
    threat_assess_score_observed: int | None = Field(
        None,
        allow_mutation=False,
        description='The Threat Assess score observed for this indicator.',
        read_only=True,
        title='threatAssessScoreObserved',
    )
    tracked_users: dict | None = Field(
        None,
        allow_mutation=False,
        description='List of tracked users and their observation and false positive stats.',
        read_only=True,
        title='trackedUsers',
    )
    type: str | None = Field(
        None,
        description='The **type** for the Indicator.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='type',
    )
    value1: str | None = Field(
        None,
        description='Custom Indicator summary field value1.',
        methods=['POST'],
        read_only=False,
        title='value1',
    )
    value2: str | None = Field(
        None,
        description='Custom Indicator summary field value2.',
        methods=['POST'],
        read_only=False,
        title='value2',
    )
    value3: str | None = Field(
        None,
        description='Custom Indicator summary field value3.',
        methods=['POST'],
        read_only=False,
        title='value3',
    )
    web_link: str | None = Field(
        None,
        allow_mutation=False,
        description='A link to the ThreatConnect details page for this entity.',
        read_only=True,
        title='webLink',
    )
    whois: dict | None = Field(
        None,
        allow_mutation=False,
        applies_to=['Host'],
        conditional_required=['Host'],
        description='The whois data for the indicator.',
        read_only=True,
        title='whois',
    )
    whois_active: bool = Field(
        None,
        applies_to=['Host'],
        description='Is whois active for the indicator?',
        methods=['POST', 'PUT'],
        read_only=False,
        title='whoisActive',
    )

    @validator('associated_artifacts', always=True, pre=True)
    @classmethod
    def _validate_artifacts(cls, v):
        if not v:
            return ArtifactsModel()  # type: ignore
        return v

    @validator('associated_cases', always=True, pre=True)
    @classmethod
    def _validate_cases(cls, v):
        if not v:
            return CasesModel()  # type: ignore
        return v

    @validator('file_actions', always=True, pre=True)
    @classmethod
    def _validate_file_actions(cls, v):
        if not v:
            return FileActionsModel()  # type: ignore
        return v

    @validator('file_occurrences', always=True, pre=True)
    @classmethod
    def _validate_file_occurrences(cls, v):
        if not v:
            return FileOccurrencesModel()  # type: ignore
        return v

    @validator('associated_groups', always=True, pre=True)
    @classmethod
    def _validate_groups(cls, v):
        if not v:
            return GroupsModel()  # type: ignore
        return v

    @validator('attributes', always=True, pre=True)
    @classmethod
    def _validate_indicator_attributes(cls, v):
        if not v:
            return IndicatorAttributesModel()  # type: ignore
        return v

    @validator('associated_indicators', 'custom_associations', always=True, pre=True)
    @classmethod
    def _validate_indicators(cls, v):
        if not v:
            return IndicatorsModel()  # type: ignore
        return v

    @validator('security_labels', always=True, pre=True)
    @classmethod
    def _validate_security_labels(cls, v):
        if not v:
            return SecurityLabelsModel()  # type: ignore
        return v

    @validator('tags', always=True, pre=True)
    @classmethod
    def _validate_tags(cls, v):
        if not v:
            return TagsModel()  # type: ignore
        return v


class IndicatorDataModel(
    BaseModel,
    title='Indicator Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Indicators Data Model"""

    data: list[IndicatorModel] | None = Field(
        [],
        description='The data for the Indicators.',
        methods=['POST', 'PUT'],
        title='data',
    )


class IndicatorsModel(
    BaseModel,
    title='Indicators Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Indicators Model"""

    _mode_support = PrivateAttr(default=True)

    data: list[IndicatorModel] | None = Field(
        [],
        description='The data for the Indicators.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# first-party
from tcex.api.tc.v3.artifacts.artifact_model import ArtifactsModel
from tcex.api.tc.v3.cases.case_model import CasesModel
from tcex.api.tc.v3.file_actions.file_action_model import FileActionsModel
from tcex.api.tc.v3.file_occurrences.file_occurrence_model import FileOccurrencesModel
from tcex.api.tc.v3.groups.group_model import GroupsModel
from tcex.api.tc.v3.indicator_attributes.indicator_attribute_model import IndicatorAttributesModel
from tcex.api.tc.v3.security_labels.security_label_model import SecurityLabelsModel
from tcex.api.tc.v3.tags.tag_model import TagsModel

# add forward references
IndicatorDataModel.update_forward_refs()
IndicatorModel.update_forward_refs()
IndicatorsModel.update_forward_refs()



"""TcEx Framework Module"""

# standard library
from datetime import datetime
from enum import Enum

# third-party
from arrow import Arrow

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql import Tql
from tcex.api.tc.v3.tql.tql_operator import TqlOperator
from tcex.api.tc.v3.tql.tql_type import TqlType


class CaseFilter(FilterABC):
    """Filter Object for Cases"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.CASES.value

    def assigned_to_user_or_group(self, operator: Enum, assigned_to_user_or_group: list | str):
        """Filter Assigned To User or Group based on **assignedToUserOrGroup** keyword.

        Args:
            operator: The operator enum for the filter.
            assigned_to_user_or_group: A value of User, Group, or None depending on the assignee.
        """
        if isinstance(assigned_to_user_or_group, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter(
            'assignedToUserOrGroup', operator, assigned_to_user_or_group, TqlType.STRING
        )

    def assignee_name(self, operator: Enum, assignee_name: list | str):
        """Filter Assignee based on **assigneeName** keyword.

        Args:
            operator: The operator enum for the filter.
            assignee_name: The user or group name assigned to the Case.
        """
        if isinstance(assignee_name, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('assigneeName', operator, assignee_name, TqlType.STRING)

    def attribute(self, operator: Enum, attribute: list | str):
        """Filter attribute based on **attribute** keyword.

        Args:
            operator: The operator enum for the filter.
            attribute: No description provided.
        """
        if isinstance(attribute, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('attribute', operator, attribute, TqlType.STRING)

    def cal_score(self, operator: Enum, cal_score: int | list):
        """Filter CalScore based on **calScore** keyword.

        Args:
            operator: The operator enum for the filter.
            cal_score: Cal score of the case.
        """
        if isinstance(cal_score, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('calScore', operator, cal_score, TqlType.INTEGER)

    def case_close_date(self, operator: Enum, case_close_date: Arrow | datetime | int | str):
        """Filter Cases Closed based on **caseCloseDate** keyword.

        Args:
            operator: The operator enum for the filter.
            case_close_date: The date/time the case was closed.
        """
        case_close_date = self.util.any_to_datetime(case_close_date).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('caseCloseDate', operator, case_close_date, TqlType.STRING)

    def case_close_time(self, operator: Enum, case_close_time: Arrow | datetime | int | str):
        """Filter Case Close Time based on **caseCloseTime** keyword.

        Args:
            operator: The operator enum for the filter.
            case_close_time: The date/time the case was closed.
        """
        case_close_time = self.util.any_to_datetime(case_close_time).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('caseCloseTime', operator, case_close_time, TqlType.STRING)

    def case_close_user(self, operator: Enum, case_close_user: list | str):
        """Filter Case Close User based on **caseCloseUser** keyword.

        Args:
            operator: The operator enum for the filter.
            case_close_user: The user who closed the case.
        """
        if isinstance(case_close_user, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('caseCloseUser', operator, case_close_user, TqlType.STRING)

    def case_detection_time(
        self, operator: Enum, case_detection_time: Arrow | datetime | int | str
    ):
        """Filter Case Detection Time based on **caseDetectionTime** keyword.

        Args:
            operator: The operator enum for the filter.
            case_detection_time: The date/time the case was detected.
        """
        case_detection_time = self.util.any_to_datetime(case_detection_time).strftime(
            '%Y-%m-%d %H:%M:%S'
        )
        self._tql.add_filter('caseDetectionTime', operator, case_detection_time, TqlType.STRING)

    def case_detection_user(self, operator: Enum, case_detection_user: list | str):
        """Filter Case Detection User based on **caseDetectionUser** keyword.

        Args:
            operator: The operator enum for the filter.
            case_detection_user: The user who logged the case detection time.
        """
        if isinstance(case_detection_user, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('caseDetectionUser', operator, case_detection_user, TqlType.STRING)

    def case_occurrence_time(
        self, operator: Enum, case_occurrence_time: Arrow | datetime | int | str
    ):
        """Filter Case Occurrence Time based on **caseOccurrenceTime** keyword.

        Args:
            operator: The operator enum for the filter.
            case_occurrence_time: The date/time the case occurred.
        """
        case_occurrence_time = self.util.any_to_datetime(case_occurrence_time).strftime(
            '%Y-%m-%d %H:%M:%S'
        )
        self._tql.add_filter('caseOccurrenceTime', operator, case_occurrence_time, TqlType.STRING)

    def case_occurrence_user(self, operator: Enum, case_occurrence_user: list | str):
        """Filter Case Occurrence User based on **caseOccurrenceUser** keyword.

        Args:
            operator: The operator enum for the filter.
            case_occurrence_user: The user who logged the case occurrence time.
        """
        if isinstance(case_occurrence_user, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('caseOccurrenceUser', operator, case_occurrence_user, TqlType.STRING)

    def case_open_date(self, operator: Enum, case_open_date: Arrow | datetime | int | str):
        """Filter Cases Created based on **caseOpenDate** keyword.

        Args:
            operator: The operator enum for the filter.
            case_open_date: The date/time the case was opened.
        """
        case_open_date = self.util.any_to_datetime(case_open_date).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('caseOpenDate', operator, case_open_date, TqlType.STRING)

    def case_open_time(self, operator: Enum, case_open_time: Arrow | datetime | int | str):
        """Filter Case Open Time based on **caseOpenTime** keyword.

        Args:
            operator: The operator enum for the filter.
            case_open_time: The date/time the case was opened.
        """
        case_open_time = self.util.any_to_datetime(case_open_time).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('caseOpenTime', operator, case_open_time, TqlType.STRING)

    def case_open_user(self, operator: Enum, case_open_user: list | str):
        """Filter Case Opening User based on **caseOpenUser** keyword.

        Args:
            operator: The operator enum for the filter.
            case_open_user: The user who opened the case.
        """
        if isinstance(case_open_user, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('caseOpenUser', operator, case_open_user, TqlType.STRING)

    def created_by(self, operator: Enum, created_by: list | str):
        """Filter Creator based on **createdBy** keyword.

        Args:
            operator: The operator enum for the filter.
            created_by: The account login of the user who created the case.
        """
        if isinstance(created_by, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('createdBy', operator, created_by, TqlType.STRING)

    def created_by_id(self, operator: Enum, created_by_id: int | list):
        """Filter Creator ID based on **createdById** keyword.

        Args:
            operator: The operator enum for the filter.
            created_by_id: The user ID for the creator of the case.
        """
        if isinstance(created_by_id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('createdById', operator, created_by_id, TqlType.INTEGER)

    def date_added(self, operator: Enum, date_added: Arrow | datetime | int | str):
        """Filter Date Added based on **dateAdded** keyword.

        Args:
            operator: The operator enum for the filter.
            date_added: The date the case was added to the system.
        """
        date_added = self.util.any_to_datetime(date_added).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('dateAdded', operator, date_added, TqlType.STRING)

    def description(self, operator: Enum, description: list | str):
        """Filter Description based on **description** keyword.

        Args:
            operator: The operator enum for the filter.
            description: The description of the case.
        """
        if isinstance(description, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('description', operator, description, TqlType.STRING)

    def detection_due(self, operator: Enum, detection_due: Arrow | datetime | int | str):
        """Filter Detection Due based on **detectionDue** keyword.

        Args:
            operator: The operator enum for the filter.
            detection_due: The date/time the case Time To Detect needs to be completed based on the
                organization SLA settings.
        """
        detection_due = self.util.any_to_datetime(detection_due).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('detectionDue', operator, detection_due, TqlType.STRING)

    def detection_overdue(self, operator: Enum, detection_overdue: bool):
        """Filter Detection Overdue based on **detectionOverdue** keyword.

        Args:
            operator: The operator enum for the filter.
            detection_overdue: A boolean flag indicating that the case has exceeded the allowed
                completion time for detection based on the organization SLA settings.
        """
        self._tql.add_filter('detectionOverdue', operator, detection_overdue, TqlType.BOOLEAN)

    @property
    def has_all_tags(self):
        """Return **TagFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.tags.tag_filter import TagFilter

        tags = TagFilter(Tql())
        self._tql.add_filter('hasAllTags', TqlOperator.EQ, tags, TqlType.SUB_QUERY)
        return tags

    @property
    def has_artifact(self):
        """Return **ArtifactFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.artifacts.artifact_filter import ArtifactFilter

        artifacts = ArtifactFilter(Tql())
        self._tql.add_filter('hasArtifact', TqlOperator.EQ, artifacts, TqlType.SUB_QUERY)
        return artifacts

    @property
    def has_case(self):
        """Return **CaseFilter** for further filtering."""
        cases = CaseFilter(Tql())
        self._tql.add_filter('hasCase', TqlOperator.EQ, cases, TqlType.SUB_QUERY)
        return cases

    @property
    def has_group(self):
        """Return **GroupFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.groups.group_filter import GroupFilter

        groups = GroupFilter(Tql())
        self._tql.add_filter('hasGroup', TqlOperator.EQ, groups, TqlType.SUB_QUERY)
        return groups

    @property
    def has_indicator(self):
        """Return **IndicatorFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.indicators.indicator_filter import IndicatorFilter

        indicators = IndicatorFilter(Tql())
        self._tql.add_filter('hasIndicator', TqlOperator.EQ, indicators, TqlType.SUB_QUERY)
        return indicators

    @property
    def has_note(self):
        """Return **NoteFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.notes.note_filter import NoteFilter

        notes = NoteFilter(Tql())
        self._tql.add_filter('hasNote', TqlOperator.EQ, notes, TqlType.SUB_QUERY)
        return notes

    @property
    def has_tag(self):
        """Return **TagFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.tags.tag_filter import TagFilter

        tags = TagFilter(Tql())
        self._tql.add_filter('hasTag', TqlOperator.EQ, tags, TqlType.SUB_QUERY)
        return tags

    @property
    def has_task(self):
        """Return **TaskFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.tasks.task_filter import TaskFilter

        tasks = TaskFilter(Tql())
        self._tql.add_filter('hasTask', TqlOperator.EQ, tasks, TqlType.SUB_QUERY)
        return tasks

    def has_workflow_template(self, operator: Enum, has_workflow_template: int | list):
        """Filter Associated Workflow Template based on **hasWorkflowTemplate** keyword.

        Args:
            operator: The operator enum for the filter.
            has_workflow_template: A nested query for association to workflow templates.
        """
        if isinstance(has_workflow_template, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter(
            'hasWorkflowTemplate', operator, has_workflow_template, TqlType.INTEGER
        )

    def id(self, operator: Enum, id: int | list):  # noqa: A002
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the case.
        """
        if isinstance(id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def id_as_string(self, operator: Enum, id_as_string: list | str):
        """Filter ID As String based on **idAsString** keyword.

        Args:
            operator: The operator enum for the filter.
            id_as_string: The ID of the case as a String.
        """
        if isinstance(id_as_string, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('idAsString', operator, id_as_string, TqlType.STRING)

    def last_updated(self, operator: Enum, last_updated: Arrow | datetime | int | str):
        """Filter Last Updated based on **lastUpdated** keyword.

        Args:
            operator: The operator enum for the filter.
            last_updated: The date the case was last updated in the system.
        """
        last_updated = self.util.any_to_datetime(last_updated).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('lastUpdated', operator, last_updated, TqlType.STRING)

    def missing_artifact_count(self, operator: Enum, missing_artifact_count: int | list):
        """Filter Missing Artifact Count For Tasks based on **missingArtifactCount** keyword.

        Args:
            operator: The operator enum for the filter.
            missing_artifact_count: Missing Artifact Count for Case Tasks.
        """
        if isinstance(missing_artifact_count, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter(
            'missingArtifactCount', operator, missing_artifact_count, TqlType.INTEGER
        )

    def name(self, operator: Enum, name: list | str):
        """Filter Name based on **name** keyword.

        Args:
            operator: The operator enum for the filter.
            name: The name of the case.
        """
        if isinstance(name, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('name', operator, name, TqlType.STRING)

    def owner(self, operator: Enum, owner: int | list):
        """Filter Owner ID based on **owner** keyword.

        Args:
            operator: The operator enum for the filter.
            owner: The Owner ID for the case.
        """
        if isinstance(owner, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('owner', operator, owner, TqlType.INTEGER)

    def owner_name(self, operator: Enum, owner_name: list | str):
        """Filter Owner Name based on **ownerName** keyword.

        Args:
            operator: The operator enum for the filter.
            owner_name: The owner name for the case.
        """
        if isinstance(owner_name, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('ownerName', operator, owner_name, TqlType.STRING)

    def resolution(self, operator: Enum, resolution: list | str):
        """Filter Resolution based on **resolution** keyword.

        Args:
            operator: The operator enum for the filter.
            resolution: The resolution of the case.
        """
        if isinstance(resolution, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('resolution', operator, resolution, TqlType.STRING)

    def response_due(self, operator: Enum, response_due: Arrow | datetime | int | str):
        """Filter Response Due based on **responseDue** keyword.

        Args:
            operator: The operator enum for the filter.
            response_due: The date/time the case Time To Respond needs to be completed based on the
                organization SLA settings.
        """
        response_due = self.util.any_to_datetime(response_due).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('responseDue', operator, response_due, TqlType.STRING)

    def response_overdue(self, operator: Enum, response_overdue: bool):
        """Filter Response Overdue based on **responseOverdue** keyword.

        Args:
            operator: The operator enum for the filter.
            response_overdue: A boolean flag indicating that the case has exceeded the allowed
                completion time for response based on the organization SLA settings.
        """
        self._tql.add_filter('responseOverdue', operator, response_overdue, TqlType.BOOLEAN)

    def severity(self, operator: Enum, severity: list | str):
        """Filter Severity based on **severity** keyword.

        Args:
            operator: The operator enum for the filter.
            severity: The severity of the case.
        """
        if isinstance(severity, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('severity', operator, severity, TqlType.STRING)

    def status(self, operator: Enum, status: list | str):
        """Filter Status based on **status** keyword.

        Args:
            operator: The operator enum for the filter.
            status: The status of the case.
        """
        if isinstance(status, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('status', operator, status, TqlType.STRING)

    def tag(self, operator: Enum, tag: list | str):
        """Filter Tag based on **tag** keyword.

        Args:
            operator: The operator enum for the filter.
            tag: The name of a tag applied to a case.
        """
        if isinstance(tag, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('tag', operator, tag, TqlType.STRING)

    def target_id(self, operator: Enum, target_id: int | list):
        """Filter Assignee ID based on **targetId** keyword.

        Args:
            operator: The operator enum for the filter.
            target_id: The assigned user or group ID for the case.
        """
        if isinstance(target_id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('targetId', operator, target_id, TqlType.INTEGER)

    def target_type(self, operator: Enum, target_type: list | str):
        """Filter Target Type based on **targetType** keyword.

        Args:
            operator: The operator enum for the filter.
            target_type: The target type for this case (either User or Group).
        """
        if isinstance(target_type, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('targetType', operator, target_type, TqlType.STRING)

    def threat_assess_score(self, operator: Enum, threat_assess_score: int | list):
        """Filter ThreatAssessScore based on **threatAssessScore** keyword.

        Args:
            operator: The operator enum for the filter.
            threat_assess_score: ThreatAssess score of the case.
        """
        if isinstance(threat_assess_score, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('threatAssessScore', operator, threat_assess_score, TqlType.INTEGER)

    def time_to_detect(self, operator: Enum, time_to_detect: int | list):
        """Filter Time To Detect based on **timeToDetect** keyword.

        Args:
            operator: The operator enum for the filter.
            time_to_detect: The amount of time (in seconds) elapsed in the Time To Detect
                calculation.
        """
        if isinstance(time_to_detect, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('timeToDetect', operator, time_to_detect, TqlType.INTEGER)

    def time_to_respond(self, operator: Enum, time_to_respond: int | list):
        """Filter Time To Respond based on **timeToRespond** keyword.

        Args:
            operator: The operator enum for the filter.
            time_to_respond: The amount of time (in seconds) elapsed in the Time To Respond
                calculation.
        """
        if isinstance(time_to_respond, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('timeToRespond', operator, time_to_respond, TqlType.INTEGER)

    def type_name(self, operator: Enum, type_name: list | str):
        """Filter Name based on **typeName** keyword.

        Args:
            operator: The operator enum for the filter.
            type_name: The name of the case.
        """
        if isinstance(type_name, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('typeName', operator, type_name, TqlType.STRING)

    def xid(self, operator: Enum, xid: list | str):
        """Filter XID based on **xid** keyword.

        Args:
            operator: The operator enum for the filter.
            xid: The XID of the case.
        """
        if isinstance(xid, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('xid', operator, xid, TqlType.STRING)



"""TcEx Framework Module"""

# third-party
from pydantic import BaseModel, Field

# first-party
from tcex.input.field_type.sensitive import Sensitive


class SmtpSettingModel(BaseModel):
    """SMTP Setting Model

    Feature: smtpSettings

    Supported for the following runtimeLevel:
    * ApiService
    * Organization
    * Playbook
    * WebhookTriggerService
    * TriggerService
    """

    tc_smtp_host: str = Field(
        None,
        description='The SMTP server hostname.',
        inclusion_reason='feature (smtpSettings)',
    )
    tc_smtp_password: Sensitive = Field(
        None,
        description='The SMTP server password.',
        inclusion_reason='feature (smtpSettings)',
    )
    tc_smtp_port: int = Field(
        None,
        description='The SMTP server port number.',
        inclusion_reason='feature (smtpSettings)',
    )
    tc_smtp_username: str = Field(
        None,
        description='The SMTP server username.',
        inclusion_reason='feature (smtpSettings)',
    )
    tc_sys_email: str = Field(
        None,
        description='The system level email address.',
        inclusion_reason='feature (smtpSettings)',
    )



"""TcEx Framework Module"""

# ruff: noqa: F401

# standard library
import contextlib
import logging

# first-party
from tcex.__metadata__ import __license__, __version__
from tcex.logger.cache_handler import CacheHandler
from tcex.logger.trace_logger import TraceLogger


def initialize_logger():
    """Initialize logger TraceLogger."""
    # init logger before instantiating tcex
    logging.setLoggerClass(TraceLogger)
    _logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore
    _logger.setLevel(logging.TRACE)  # type: ignore

    # add TEMP cache handler, which will be removed in tcex.py (we don't know log path here)
    cache = CacheHandler()
    cache.set_name('cache')
    cache.setLevel(logging.TRACE)  # type: ignore
    cache.setFormatter(
        logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)8s - %(message)s '
            '(%(filename)s:%(funcName)s:%(lineno)d)'
        )
    )
    _logger.addHandler(cache)


# init logger before instantiating tcex
initialize_logger()

with contextlib.suppress(ImportError):
    # first-party
    from tcex.tcex import TcEx



"""TcEx Framework Module"""

# standard library
import json
import logging
from collections import OrderedDict
from functools import cached_property
from pathlib import Path
from typing import Any

from .install_json_update import InstallJsonUpdate
from .install_json_validate import InstallJsonValidate
from .model.install_json_model import InstallJsonModel, ParamsModel

# get logger
_logger = logging.getLogger(__name__.split('.', maxsplit=1)[0])


class InstallJson:
    """Config object for install.json configuration file

    This class can't be a Singleton because it's used by the package
    command to create a install.json file in the build directory.
    """

    def __init__(
        self,
        filename: str | None = None,
        path: Path | str | None = None,
        logger: logging.Logger | None = None,
    ):
        """Initialize instance properties."""
        filename = filename or 'install.json'
        path = Path(path or Path.cwd())
        self.log = logger or _logger

        # properties
        self.fqfn = path / filename

    @property
    def app_prefix(self) -> str:
        """Return the appropriate output var type for the current App."""
        return self.app_prefixes.get(self.model.runtime_level.lower(), '')

    @property
    def app_prefixes(self) -> dict[str, str]:
        """Return all the current App prefixes."""
        return {
            'organization': 'TC_-_',
            'playbook': 'TCPB_-_',
            'apiservice': 'TCVA_-_',
            'feedapiservice': 'TCVF_-_',
            'triggerservice': 'TCVC_-_',
            'webhooktriggerservice': 'TCVW_-_',
        }

    @cached_property
    def contents(self) -> dict:
        """Return install.json file contents."""
        if self.fqfn.is_file():
            try:
                with self.fqfn.open() as fh:
                    return json.load(fh, object_pairs_hook=OrderedDict)
            except (OSError, ValueError):  # pragma: no cover
                self.log.exception(
                    f'feature=install-json, exception=failed-reading-file, filename={self.fqfn}'
                )

        return {
            'displayName': 'External App',
            'features': [],
            'languageVersion': '3.11',
            'listDelimiter': '|',
            'programLanguage': 'python',
            'programMain': 'run.py',
            'programVersion': '0.0.0',
            'runtimeLevel': 'external',
            'sdkVersion': '4.0.0',
        }

    def create_output_variables(self, output_variables: list, job_id: int = 9876) -> list:
        """Create output variables.

        # "#App:9876:app.data.count!String"
        # "#Trigger:9876:app.data.count!String"

        Args:
            output_variables: A dict of the output variables
            job_id: A job id to use in output variable string.
        """
        variables = []
        for p in output_variables:
            variables.append(self.create_variable(p.name, p.type, job_id))
        return variables

    def create_variable(self, var_name: str, var_type: str, job_id: int = 1234) -> str:
        """Create output variables.

        # "#App:9876:app.data.count!String"
        # "#Trigger:9876:app.data.count!String"

        Args:
            var_name: The variable name.
            var_type: The variable type.
            job_id: A job id to use in output variable string.
        """
        return f'#{self.model.app_output_var_type}:{job_id}:{var_name}!{var_type}'

    @staticmethod
    def expand_valid_values(valid_values: list) -> list[str]:
        """Expand supported playbook variables to their full list.

        Args:
            valid_values: The list of valid values for Choice or MultiChoice inputs.

        Returns:
            list: An expanded list of valid values for Choice or MultiChoice inputs.
        """
        # expand is typically for Choice inputs and
        # ${FILE} is usually only on String inputs
        # if '${FILE}' in valid_values:
        #     valid_values.remove('${FILE}')

        # valid_values = list(valid_values)
        if '${GROUP_TYPES}' in valid_values:
            valid_values.remove('${GROUP_TYPES}')
            valid_values.extend(
                [
                    'Adversary',
                    'Attack Pattern',
                    'Campaign',
                    'Course of Action',
                    'Document',
                    'Email',
                    'Event',
                    'Incident',
                    'Intrusion Set',
                    'Malware',
                    'Signature',
                    'Tactic',
                    'Task',
                    'Tool',
                    'Threat',
                    'Vulnerability',
                ]
            )

        if '${OWNERS}' in valid_values:
            valid_values.remove('${OWNERS}')

        if '${USERS}' in valid_values:
            valid_values.remove('${USERS}')
        return valid_values

    def has_feature(self, feature: str) -> bool:
        """Return True if App has the provided feature."""
        return feature.lower() in [f.lower() for f in self.model.features]

    @cached_property
    def is_external_app(self) -> bool:
        """Return True if App does not have a install.json file."""
        return not self.fqfn.is_file()

    # this needs to be a cached property, for tcex package to update install.json properly
    @cached_property
    def model(self) -> InstallJsonModel:
        """Return the Install JSON model."""
        return InstallJsonModel(**self.contents)

    @property
    def params_dict(self) -> dict[str, ParamsModel]:
        """Return params as name/model.

        Used in tcex_testing for dynamic generation of output variables.
        """
        params = {}
        for p in self.model.params:
            params.setdefault(p.name, p)
        return params

    def params_to_args(
        self,
        name: str | None = None,
        hidden: bool | None = None,
        required: bool | None = None,
        service_config: bool | None = None,
        _type: str | None = None,
        input_permutations: dict | None = None,
    ) -> dict[str, Any]:
        """Return params as cli args.

        Used by tcex_testing project.

        Args:
            name: The name of the input to return.
            hidden: If set the inputs will be filtered based on hidden field.
            required: If set the inputs will be filtered based on required field.
            service_config: If set the inputs will be filtered based on serviceConfig field.
            _type: The type of input to return.
            input_permutations: A list of valid input names for provided permutation.

        Returns:
            dict: All args for current filter
        """
        args = {}
        for n, p in self.model.filter_params(
            name, hidden, required, service_config, _type, input_permutations
        ).items():
            if p.type.lower() == 'boolean':
                args[n] = p.default
            elif p.type.lower() == 'choice':
                # provide all options by default
                valid_values = f'[{"|".join(self.expand_valid_values(p.valid_values))}]'

                # use the value from input_permutations if available or provide valid values
                if input_permutations is not None:
                    valid_values = input_permutations.get(n, valid_values)

                # use default if available
                if p.default is not None:
                    valid_values = p.default

                args[n] = valid_values
            elif p.type.lower() == 'multichoice':
                args[n] = p.valid_values
            elif p.type.lower() == 'keyvaluelist':
                args[n] = '<KeyValueArray>'
            elif n in ['tc_api_access_id', 'tc_api_secret_key']:  # pragma: no cover
                # leave these parameters set to the value defined in defaults
                pass
            else:
                types = '|'.join(p.playbook_data_type)
                if types:
                    args[n] = p.default or f'<{types}>'
                else:  # pragma: no cover
                    args[n] = p.default or ''
        return args

    @property
    def tc_playbook_out_variables(self) -> list:
        """Return playbook output variable name array"""
        if self.model.playbook is None:
            return []
        return self.create_output_variables(self.model.playbook.output_variables)

    @property
    def tc_playbook_out_variables_csv(self) -> str:
        """Return playbook output variables as CSV string"""
        return ','.join(self.tc_playbook_out_variables)

    @property
    def update(self) -> InstallJsonUpdate:
        """Return InstallJsonUpdate instance."""
        return InstallJsonUpdate(ij=self)

    @property
    def validate(self) -> InstallJsonValidate:
        """Validate install.json."""
        return InstallJsonValidate(ij=self)

    def write(self):
        """Write current data file."""
        data = self.model.json(
            by_alias=True, exclude_defaults=True, exclude_none=True, indent=2, sort_keys=True
        )
        with self.fqfn.open(mode='w') as fh:
            fh.write(f'{data}\n')



"""TcEx Framework Module"""

# standard library
import logging
from abc import ABC
from collections.abc import Generator
from typing import Self

# third-party
from requests import Response, Session
from requests.exceptions import ProxyError, RetryError

# first-party
from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC
from tcex.api.tc.v3.tql.tql_operator import TqlOperator
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.exit.error_code import handle_error
from tcex.logger.trace_logger import TraceLogger
from tcex.pleb.cached_property import cached_property
from tcex.util import Util

# get tcex logger
_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore


class ObjectABC(ABC):  # noqa: B024
    """Object Abstract Base Class

    This class is a base class for Object classes that use
    multi-inheritance with a pydantic BaseModel class. To ensure
    properties are not added to the model both @property and @setter
    methods are used.
    """

    def __init__(self, session: Session):
        """Initialize instance properties."""
        self._session: Session = session

        # properties
        self._parent_data = {}
        self._remove_objects = {
            'associations': [],
            'attributes': [],
            'securityLabels': [],
            'tags': [],
        }
        self.util = Util()
        self.log = _logger
        self.request: Response

        # define/overwritten in child class
        self._model: V3ModelABC
        self._nested_field_name: str | None = None
        self._nested_filter: str | None = None
        self.type_: str

    @property
    def _api_endpoint(self) -> str:  # pragma: no cover
        """Return the type specific API endpoint."""
        ex_msg = 'Child class must implement this property.'
        raise NotImplementedError(ex_msg)

    def _calculate_unique_id(self) -> dict[str, int | str]:
        if self.model.id:
            return {'filter': 'id', 'value': self.model.id}

        if hasattr(self.model, 'xid') and self.model.xid:  # type: ignore
            return {'filter': 'xid', 'value': self.model.xid}  # type: ignore

        if self.type_.lower() in ['indicator']:
            return {'filter': 'summary', 'value': self.model.summary}  # type: ignore

        return {}

    def _iterate_over_sublist(
        self, sublist_type: ObjectCollectionABC, custom_associations: bool = False
    ) -> Generator[Self, None, None]:
        """Iterate over any nested collections."""
        sublist = sublist_type(session=self._session)  # type: ignore

        # determine the filter type and value based on the available object fields.
        unique_id_data = self._calculate_unique_id()

        # id!=2984993+AND+hasGroup(typename="all"+AND+isGroup+=+false+AND+hasIndicator(id=2984993))
        # add the filter (e.g., group.has_indicator.id(TqlOperator.EQ, 123)) for the parent object.
        if custom_associations is True:
            sublist.filter.tql = (
                f'id!={self.model.id} AND hasGroup(typename="all" '
                f'AND isGroup = false AND hasIndicator(id={self.model.id}))'
            )
            sublist.params = {'fields': ['associationName']}
        else:
            getattr(
                getattr(sublist.filter, self._nested_filter),  # type: ignore
                unique_id_data.get('filter'),  # type: ignore
            )(TqlOperator.EQ, unique_id_data.get('value'))

        # return the sub object, injecting the parent data
        for obj in sublist:
            obj._parent_data = {  # noqa: SLF001
                'api_endpoint': self._api_endpoint,
                'type': self.type_,
                'unique_id': unique_id_data.get('value'),
            }
            yield obj
        self.request = sublist.request

    def _request(
        self,
        method: str,
        url: str,
        body: bytes | str | None = None,
        params: dict | None = None,
        headers: dict | None = None,
    ):
        """Handle standard request with error checking."""
        max_length = 1_000
        try:
            self.request = self._session.request(
                method, url, data=body, headers=headers, params=params
            )
            if (
                isinstance(self.request.request.body, str)
                and len(self.request.request.body) < max_length
            ):
                self.log.debug(f'feature=api-tc-v3, request-body={self.request.request.body}')
        except (ConnectionError, ProxyError, RetryError):  # pragma: no cover
            handle_error(
                code=951,
                message_values=[
                    method.upper(),
                    None,
                    '{"message": "Connection/Proxy Error/Retry"}',
                    url,
                ],
            )

        content_type = self.request.headers.get('Content-Type')
        if content_type == 'application/json' and not self.success(self.request):
            err = self.request.text or self.request.reason
            handle_error(
                code=952,
                message_values=[
                    (self.request.request.method or '').upper(),
                    self.request.status_code,
                    err,
                    self.request.url,
                ],
            )

        # log content for debugging
        if content_type == 'application/json':
            self.log_response_text(self.request)

    @staticmethod
    def _validate_id(id_: int | str | None, url: str):
        """Raise exception is id is not provided."""
        if not id_:  # pragma: no cover
            message = '{"message": "No ID provided.", "status": "Error"}'
            handle_error(code=952, message_values=['GET', '404', message, url])

    @property
    def as_entity(self) -> dict[str, str]:  # pragma: no cover
        """Return the entity representation of the object."""
        ex_msg = 'Child class must implement this property.'
        raise NotImplementedError(ex_msg)

    @property
    def available_fields(self) -> list[str]:
        """Return the available query param field names for this object."""
        return [fd['name'] for fd in self.fields]

    def create(self, params: dict | None = None) -> Response:
        """Create or Update the Case Management object.

        This is determined based on if the id is already present in the object.
        """
        method = 'POST'
        body = self.model.gen_body_json(method=method)
        params = self.gen_params(params) if params else None
        self._request(
            method,
            self.url(method),
            body,
            headers={'content-type': 'application/json'},
            params=params,
        )

        # get the response data from nested data object or full response
        response_json = self.request.json()

        # update the model with the response from the API
        self.model = type(self.model)(**response_json.get('data'))

        return self.request

    def delete(self, params: dict | None = None):
        """Delete the object."""
        method = 'DELETE'
        body = self.model.gen_body_json(method)

        # get the unique id value for id, xid, summary, etc ...
        unique_id = self._calculate_unique_id().get('value')

        # validate an id is available
        self._validate_id(unique_id, self.url(method, unique_id))

        self._request(method, self.url(method, unique_id), body, params=params)

        return self.request

    @cached_property
    def fields(self) -> list[dict[str, str]]:
        """Return the field data for this object."""
        _fields = []
        r = self._session.options(f'{self._api_endpoint}/fields', params={})
        if r.ok:
            _fields = r.json().get('data', [])
        return _fields

    def gen_params(self, params: dict) -> dict:
        """Return appropriate params values."""
        # convert all keys to camel case
        params = {self.util.snake_to_camel(k): v for k, v in params.items()}

        # special parameter for indicators to enable the return the the indicator fields
        # (value1, value2, value3) on std-custom/custom-custom indicator types.
        if self.type_ == 'Indicator':
            params.setdefault('fields', []).append('genericCustomIndicatorValues')

        # add fields parameter if provided
        if '_all_' in params.get('fields', []):
            params['fields'] = list(self.available_fields)

        return params

    def get(
        self,
        object_id: int | None = None,
        params: dict | None = None,
    ) -> Response:
        """Get the Case Management Object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': [
                    'caseId',
                    'summary',
                ],  # Additional fields returned on the results
            }

        Args:
            object_id: The unique id of the object to be returned.
            params: Dict of the params to be sent while retrieving the Artifacts objects.
        """
        method = 'GET'
        object_id = object_id or self.model.id
        params = self.gen_params(params) if params else None

        # get the unique id value for id, xid, summary, etc ...
        unique_id = self._calculate_unique_id().get('value')

        # validate an id is available
        self._validate_id(unique_id, self.url(method, unique_id))

        body = self.model.gen_body_json(method)
        self._request(method, self.url(method, unique_id), body, params)

        # update model
        self.model = self.request.json().get('data')

        return self.request

    def log_response_text(self, response: Response):
        """Log the response text."""
        max_text_length = 5_000
        response_text = 'response text: (text to large to log)'
        if len(response.content) < max_text_length:  # check size of content for performance
            response_text = response.text
        self.log.debug(f'feature=api-tc-v3, response-body={response_text}')

    @property
    def model(self) -> V3ModelABC:
        """Return the model data."""
        return self._model

    @model.setter
    def model(self, data: dict | V3ModelABC):
        """Create model using the provided data."""
        if isinstance(data, type(self.model)):
            # provided data is already a model, nothing required to change
            self._model = data
        elif isinstance(data, dict):
            # provided data is raw response, load the model
            self._model = type(self.model)(**data)
        else:
            ex_msg = f'Invalid data type: {type(data)} provided.'
            raise RuntimeError(ex_msg)  # noqa: TRY004

    @cached_property
    def properties(self) -> dict[str, dict | list]:
        """Return defined API properties for the current object.

        This property is used in testing API consistency.
        """
        _properties = {}
        try:
            r = self._session.options(
                self._api_endpoint,
                params={'show': 'readOnly'},
                headers={'content-type': 'application/json'},
            )
            if r.ok:
                _properties = r.json()
        except (ConnectionError, ProxyError):
            handle_error(
                code=951,
                message_values=[
                    'OPTIONS',
                    407,
                    '{"message": "Connection Error"}',
                    self._api_endpoint,
                ],
            )
        return _properties

    @staticmethod
    def success(r: Response) -> bool:
        """Validate the response is valid.

        Args:
            r: The response object.

        Returns:
            bool: True if status is "ok"
        """
        status = True
        if r.ok:
            try:
                if r.json().get('status') != 'Success':  # pragma: no cover
                    status = False
            except Exception:  # pragma: no cover
                status = False
        else:
            status = False
        return status

    def update(self, mode: str | None = None, params: dict | None = None) -> Response:
        """Create or Update the Case Management object.

        This is determined based on if the id is already present in the object.
        """
        method = 'PUT'
        body = self.model.gen_body_json(method=method, mode=mode)
        params = self.gen_params(params) if params else None

        # get the unique id value for id, xid, summary, etc ...
        unique_id = self._calculate_unique_id().get('value')

        # validate an id is available
        self._validate_id(unique_id, self.url(method))

        self._request(
            method,
            self.url(method, unique_id=unique_id),
            body,
            headers={'content-type': 'application/json'},
            params=params,
        )

        # get the response data from nested data object or full response
        response_json = self.request.json()

        self.model = type(self.model)(**response_json.get('data'))

        return self.request

    def url(self, method: str, unique_id: int | str | None = None) -> str:
        """Return the proper URL."""
        unique_id = unique_id or self._calculate_unique_id().get('value')
        if method in ['DELETE', 'GET', 'PUT']:
            return f'{self._api_endpoint}/{unique_id}'
        return self._api_endpoint



"""TcEx Framework Module"""

# third-party
from pydantic import BaseModel, Field

# first-party
from tcex.input.field_type.sensitive import Sensitive


class ProxyModel(BaseModel):
    """Proxy Model

    Supported for the following runtimeLevel:
    * ApiService
    * Playbook
    * Organization
    * WebhookTriggerService
    * TriggerService
    """

    tc_proxy_host: str | None = Field(
        None,
        description='The proxy hostname.',
        inclusion_reason='runtimeLevel',
    )
    tc_proxy_port: int | None = Field(
        None,
        description='The proxy port number.',
        inclusion_reason='runtimeLevel',
    )
    tc_proxy_username: str | None = Field(
        None,
        description='The proxy username.',
        inclusion_reason='runtimeLevel',
    )
    tc_proxy_password: Sensitive | None = Field(
        None,
        description='The proxy password',
        inclusion_reason='runtimeLevel',
    )
    tc_proxy_external: bool = Field(
        default=False,
        description='Flag to enable proxy for external connections.',
        inclusion_reason='runtimeLevel',
    )
    tc_proxy_tc: bool = Field(
        default=False,
        description='Flag to enable proxy for ThreatConnect connection.',
        inclusion_reason='runtimeLevel',
    )



"""TcEx Framework Module"""

# standard library
import json


class Association:
    """ThreatConnect Batch Indicator Object"""

    __slots__ = ['association_type', 'id_1', 'id_2', 'ref_1', 'ref_2', 'type_1', 'type_2']

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        self.association_type = kwargs.get('association_type') or kwargs.get('associationType')
        self.ref_1 = kwargs.get('ref_1')
        self.type_1 = kwargs.get('type_1')
        self.id_1 = kwargs.get('id_1')
        self.ref_2 = kwargs.get('ref_2')
        self.type_2 = kwargs.get('type_2')
        self.id_2 = kwargs.get('id_2')

    @property
    def data(self) -> dict:
        """Return Indicator data."""
        # add attributes
        association = {
            'associationType': self.association_type,
            'ref_1': self.ref_1,
            'type_1': self.type_1,
            'id_1': self.id_1,
            'ref_2': self.ref_2,
            'type_2': self.type_2,
            'id_2': self.id_2,
        }
        return {k: v for k, v in association.items() if v is not None}

    def _normalized_data(self) -> dict:
        """Return normalized data for the association."""
        # Combine the ref, type, and id values into tuples
        combined = [(self.ref_1, self.type_1, self.id_1), (self.ref_2, self.type_2, self.id_2)]
        # Sort the combined tuples
        sorted_combined = sorted(combined, key=lambda x: (x[0] is None, x[0], x[1], x[2]))
        # Create the association dictionary using the sorted values
        association = {
            'associationType': self.association_type,
            'ref_1': sorted_combined[0][0],
            'type_1': sorted_combined[0][1],
            'id_1': sorted_combined[0][2],
            'ref_2': sorted_combined[1][0],
            'type_2': sorted_combined[1][1],
            'id_2': sorted_combined[1][2],
        }
        # Remove any None values from the dictionary
        return {k: v for k, v in association.items() if v is not None}

    def __eq__(self, other):
        """Check if two Association objects are equal."""
        if not isinstance(other, Association):
            return NotImplemented
        return self._normalized_data() == other._normalized_data()

    def __hash__(self):
        """Return hash for the object."""
        return hash(tuple(self._normalized_data().items()))

    def __str__(self) -> str:
        """Return string representation of the object."""
        return json.dumps(self.data, indent=4)



"""Declares a scoped_property decorator."""

# standard library
import os
import threading
from collections.abc import Callable
from typing import Any, ClassVar, Generic, TypeVar

T = TypeVar('T')


class scoped_property(Generic[T]):  # noqa: N801
    """Makes a value unique for each thread and also acts as a @property decorator.

    Essentially, a thread-and-process local value.  When used to decorate a function, will
    treat that function as a factory for the underlying value, and will invoke it to produce a value
    for each thread the value is requested from.

    Note that this also provides a cache: each thread will reuse the value previously created
    for it.
    """

    instances: ClassVar = []

    def __init__(self, wrapped: Callable[..., T]):
        """Initialize the instance properties."""
        scoped_property.instances.append(self)
        self.wrapped = wrapped
        self.value = threading.local()

    def __del__(self):
        """Remove instance from the instances class variable when it's destroyed."""
        scoped_property.instances = [i for i in scoped_property.instances if i != self]

    def __get__(self, instance: Any, _: Any) -> T:
        """Return a thread-and-process-local value.

        Implementation per the descriptor protocol.

        Args:
            instance: the instance this property is being resolved for.
            owner: same as instance.
        """
        if hasattr(self.value, 'data'):
            # A value has been created for this thread already, but we have to make sure we're in
            # the same process (threads are duplicated when a process is forked).
            pid, value, stored_instance = self.value.data

            # create a new instance if the following are not true:
            # 1. Check for same pid
            # 2. Check to ensure stored instance is the same as the current instance
            if pid != os.getpid() or stored_instance is not instance:
                return self._create_value(instance, self.wrapped)

            return value

        # A value has *not* been created for the calling thread
        # yet, so use the factory to create a new one.
        return self._create_value(instance, self.wrapped)

    def _create_value(self, instance, wrapped, *args, **kwargs) -> T:
        """Call the wrapped factory function to get a new value."""
        data = wrapped(instance, *args, **kwargs) if instance else wrapped(*args, **kwargs)

        # add data to threat.local
        self.value.data = os.getpid(), data, instance

        return data

    @staticmethod
    def _reset():
        for i in scoped_property.instances:
            i.value = threading.local()



"""TcEx Framework Module"""

# standard library
import os
import threading
from pathlib import Path
from typing import TYPE_CHECKING

# first-party
from tcex.app.config.install_json import InstallJson
from tcex.app.key_value_store.key_value_store import KeyValueStore
from tcex.app.playbook.playbook import Playbook
from tcex.app.token import Token
from tcex.input.model.module_app_model import ModuleAppModel
from tcex.pleb.cached_property import cached_property
from tcex.pleb.scoped_property import scoped_property
from tcex.registry import registry
from tcex.util.file_operation import FileOperation

if TYPE_CHECKING:
    # first-party
    from tcex.app.service import ApiService, CommonServiceTrigger, WebhookTriggerService


class App:
    """TcEx Module"""

    def __init__(self, model: ModuleAppModel, proxies: dict[str, str]):
        """Initialize instance properties."""
        self.model = model
        self.proxies = proxies

    @cached_property
    def file_operation(self) -> FileOperation:
        """Include the Utils module."""
        return FileOperation(
            out_path=self.model.tc_out_path,
            temp_path=self.model.tc_temp_path,
        )

    def get_playbook(
        self, context: str | None = None, output_variables: list | None = None
    ) -> Playbook:
        """Return a new instance of playbook module.

        Args:
            context: The KV Store context/session_id. For PB Apps the context is provided on
                startup, but for service Apps each request gets a different context.
            output_variables: The requested output variables. For PB Apps outputs are provided on
                startup, but for service Apps each request gets different outputs.
        """
        return Playbook(self.key_value_store, context, output_variables)

    @scoped_property
    def key_value_store(self) -> KeyValueStore:
        """Return the correct KV store for this execution.

        The TCKeyValueAPI KV store is limited to two operations (create and read),
        while the Redis kvstore wraps a few other Redis methods.
        """
        return KeyValueStore(
            registry.session_tc,
            self.model.tc_kvstore_host,
            self.model.tc_kvstore_port,
            self.model.tc_kvstore_type,
            self.model.tc_playbook_kvstore_id,
            self.model.tc_kvstore_pass,
            self.model.tc_kvstore_user,
            self.model.tc_kvstore_tls_enabled,
            self.model.tc_kvstore_tls_port,
            self.model.tc_svc_broker_cacert_file,
            self.model.tc_svc_broker_cert_file,
            self.model.tc_svc_broker_key_file,
        )

    @cached_property
    def ij(self) -> InstallJson:
        """Return the install.json file as a dict."""
        return self.install_json

    @cached_property
    def install_json(self) -> InstallJson:
        """Return the install.json file as a dict."""
        return InstallJson()

    @scoped_property
    def playbook(self) -> Playbook:
        """Return an instance of Playbooks module.

        This property defaults context and output variables to arg values.
        """
        return self.get_playbook(
            context=self.model.tc_playbook_kvstore_context,
            output_variables=self.model.tc_playbook_out_variables,
        )

    def results_tc(self, key: str, value: str):
        """Write data to results_tc file in TcEX specified directory.

        The TcEx platform support persistent values between executions of the App.  This
        method will store the values for TC to read and put into the Database.

        Args:
            key: The data key to be stored.
            value: The data value to be stored.
        """
        if os.access(self.model.tc_out_path, os.W_OK):
            results_file = Path(self.model.tc_out_path) / 'results.tc'
        else:
            results_file = Path('results.tc')

        new = True
        results_file.open('a').close()  # create file
        with results_file.open('r+') as fh:
            results = ''
            for line in fh.read().strip().split('\n'):
                if not line:
                    continue
                try:
                    k, v = line.split(' = ')
                except ValueError:
                    # handle null/empty value (e.g., "name =")
                    k, v = line.split(' =')
                if k == key:
                    v = value
                    new = False
                if v is not None:
                    results += f'{k} = {v}\n'
            if new and value is not None:  # indicates the key/value pair didn't already exist
                results += f'{key} = {value}\n'
            fh.seek(0)
            fh.write(results)
            fh.truncate()

    @cached_property
    def service(self) -> 'ApiService | CommonServiceTrigger | WebhookTriggerService':
        """Include the Service Module."""
        if self.install_json.model.is_api_service_app:
            # first-party
            from tcex.app.service import ApiService as Service
        elif self.ij.model.is_webhook_trigger_app:
            # first-party
            from tcex.app.service import WebhookTriggerService as Service
        elif self.ij.model.is_trigger_app:
            # first-party
            from tcex.app.service import CommonServiceTrigger as Service
        else:
            ex_msg = 'Could not determine the service type.'
            raise RuntimeError(ex_msg)

        return Service(self.key_value_store, registry.logger, self.model, self.token)

    @cached_property
    def token(self) -> Token:
        """Return token object."""
        _proxies = None
        if self.model.tc_proxy_tc is True:
            _proxies = self.proxies

        _tokens = Token(
            self.model.tc_api_path,
            self.model.tc_verify,
            _proxies,
        )

        # register token for Apps that pass token on start
        if all([self.model.tc_token, self.model.tc_token_expires]):
            _tokens.register_token(
                key=threading.current_thread().name,
                token=self.model.tc_token,
                expires=self.model.tc_token_expires,
            )
        return _tokens

    @cached_property
    def user_agent(self) -> dict[str, str]:
        """Return a User-Agent string."""
        return {
            'User-Agent': (
                f'TcEx/{__import__(__name__).__version__}, '
                f'{self.ij.model.display_name}/{self.ij.model.program_version}'
            )
        }



"""TcEx Framework Module"""

# standard library
import json
from collections.abc import Iterator

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.object_abc import ObjectABC
from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC
from tcex.api.tc.v3.tags.tag_filter import TagFilter
from tcex.api.tc.v3.tags.tag_model import TagModel, TagsModel


class Tag(ObjectABC):
    """Tags Object.

    Args:
        description (str, kwargs): A brief description of the Tag.
        name (str, kwargs): The **name** for the Tag.
        owner (str, kwargs): The name of the Owner of the Tag.
        security_coverage (object, kwargs): For ATT&CK-based tags, this is the security coverage
            level assigned to the tag.
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(kwargs.pop('session', None))

        # properties
        self._model: TagModel = TagModel(**kwargs)
        self._nested_field_name = 'tags'
        self._nested_filter = 'has_tag'
        self.type_ = 'Tag'

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.TAGS.value

    @property
    def model(self) -> TagModel:
        """Return the model data."""
        return self._model

    @model.setter
    def model(self, data: dict | TagModel):
        """Create model using the provided data."""
        if isinstance(data, type(self.model)):
            # provided data is already a model, nothing required to change
            self._model = data
        elif isinstance(data, dict):
            # provided data is raw response, load the model
            self._model = type(self.model)(**data)
        else:
            ex_msg = f'Invalid data type: {type(data)} provided.'
            raise RuntimeError(ex_msg)  # noqa: TRY004

    def remove(self, params: dict | None = None):
        """Remove a nested object."""
        method = 'PUT'
        unique_id = self._calculate_unique_id()

        # validate an id is available
        self._validate_id(unique_id.get('value'), '')

        body = json.dumps(
            {
                self._nested_field_name: {
                    'data': [{unique_id.get('filter'): unique_id.get('value')}],
                    'mode': 'delete',
                }
            }
        )

        # get the unique id value for id, xid, summary, etc ...
        parent_api_endpoint = self._parent_data.get('api_endpoint')
        parent_unique_id = self._parent_data.get('unique_id')
        url = f'{parent_api_endpoint}/{parent_unique_id}'

        # validate parent an id is available
        self._validate_id(parent_unique_id, url)

        self._request(
            method=method,
            url=url,
            body=body,
            headers={'content-type': 'application/json'},
            params=params,
        )

        return self.request


class Tags(ObjectCollectionABC):
    """Tags Collection.

    # Example of params input
    {
        "result_limit": 100,  # Limit the retrieved results.
        "result_start": 10,  # Starting count used for pagination.
        "fields": ["caseId", "summary"]  # Select additional return fields.
    }

    Args:
        session (Session): Session object configured with TC API Auth.
        tql_filters (list): List of TQL filters.
        params (dict): Additional query params (see example above).
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(
            kwargs.pop('session', None), kwargs.pop('tql_filter', None), kwargs.pop('params', None)
        )
        self._model = TagsModel(**kwargs)
        self.type_ = 'tags'

    def __iter__(self) -> Iterator[Tag]:
        """Return CM objects."""
        return self.iterate(base_class=Tag)  # type: ignore

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.TAGS.value

    @property
    def filter(self) -> TagFilter:
        """Return the type specific filter object."""
        return TagFilter(self.tql)



"""TcEx Framework Module"""

# standard library
import atexit
from functools import cached_property

# third-party
import redis


class RedisClientSsl:
    """A shared REDIS client connection using a Connection Pool.

    Initialize a single shared redis.connection.ConnectionPool.
    For a full list of kwargs see https://redis-py.readthedocs.io/en/latest/#redis.Connection.

    Args:
        host: The Redis host. Defaults to localhost.
        port: The Redis port. Defaults to 6379.
        db: The Redis db. Defaults to 0.
        username (str, kwargs): The REDIS username.
        password (str, kwargs): The REDIS password.
        ssl_certfile (str, kwargs): The REDIS SSL certfile.
        ssl_keyfile (str, kwargs): The REDIS SSL keyfile.
        errors (str, kwargs): The REDIS errors policy (e.g. strict).
        max_connections (int, kwargs): The maximum number of connections to REDIS.
        socket_timeout (int, kwargs): The REDIS socket timeout.
        timeout (int, kwargs): The REDIS Blocking Connection Pool timeout value.
    """

    def __init__(
        self,
        host: str = 'localhost',
        port: int = 6379,
        db: int = 0,
        username: str | None = None,
        password: str | None = None,
        ssl_ca_certs: str | None = None,
        ssl_certfile: str | None = None,
        ssl_keyfile: str | None = None,
        **kwargs,
    ):
        """Initialize class properties

        Redis Client wrapper will support 4 modes:
        1. Redis Connection Pool non-secured
        2. Redis Connection Pool secured
        3. Redis Blocking Connection Pool non-secured
        4. Redis Blocking Connection Pool secured
        """
        self.pool = redis.ConnectionPool(
            connection_class=redis.SSLConnection,
            db=db,
            host=host,
            password=password,
            port=port,
            username=username,
            ssl_cert_reqs='required',
            ssl_ca_certs=ssl_ca_certs,
            ssl_certfile=ssl_certfile,
            ssl_keyfile=ssl_keyfile,
            **kwargs,
        )
        atexit.register(self.pool.disconnect)

    @cached_property
    def client(self) -> redis.Redis:
        """Return an instance of redis.client.Redis."""
        client = redis.Redis(connection_pool=self.pool)
        atexit.register(client.close)
        return client



"""TcEx Framework Module"""

# standard library
from collections import OrderedDict

# third-party
from pydantic import BaseModel, Field
from pydantic.types import constr

from ....pleb.none_model import NoneModel

__all__ = ['LayoutJsonModel']


def snake_to_camel(snake_string: str) -> str:
    """Convert snake_case to camelCase"""
    components = snake_string.split('_')
    return components[0] + ''.join(x.title() for x in components[1:])


class ParametersModel(BaseModel):
    """Model definition for layout_json.inputs.{}"""

    display: str | None
    name: str

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True


class InputsModel(BaseModel):
    """Model definition for layout_json.inputs"""

    parameters: list[ParametersModel]
    sequence: int
    title: constr(min_length=3, max_length=100)  # type: ignore

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True


class OutputsModel(BaseModel):
    """Model definition for layout_json.outputs"""

    display: str | None
    name: str

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True


class LayoutJsonModel(BaseModel):
    """Model definition for layout.json configuration file"""

    inputs: list[InputsModel]
    outputs: list[OutputsModel] = Field([], description='Layout output variable definitions.')

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True

    def get_param(self, name: str) -> NoneModel | ParametersModel:
        """Return the param or a None Model."""
        return self.params.get(name) or NoneModel()

    def get_output(self, name: str) -> NoneModel | OutputsModel:
        """Return layout.json outputs in a flattened dict with name param as key."""
        return self.outputs_.get(name) or NoneModel()

    @property
    def outputs_(self) -> dict[str, OutputsModel]:
        """Return layout.json outputs in a flattened dict with name param as key."""
        return {o.name: o for o in self.outputs}

    @property
    def param_names(self) -> list:
        """Return all param names in a single list."""
        return list(self.params.keys())

    @property
    def params(self) -> dict[str, ParametersModel]:
        """Return layout.json params in a flattened dict with name param as key."""
        # return {p.name: p for i in self.inputs for p in i.parameters}

        # order is required for display clauses to be evaluated correctly
        parameters = OrderedDict()  # remove after python 3.7
        for i in self.inputs:
            for p in i.parameters:
                parameters.setdefault(p.name, p)
        return parameters


OutputsModel.update_forward_refs()



"""TcEx Framework Module"""

# third-party
from requests import Session  # TYPE-CHECKING

# first-party
from tcex.api.tc.tc import TC
from tcex.input.input import Input  # TYPE-CHECKING
from tcex.pleb.cached_property import cached_property


class API:
    """API

    Args:
        inputs: An instance of the Input class.
        session_tc: An configured instance of request.Session with TC API Auth.
    """

    def __init__(self, inputs: Input, session_tc: Session):
        """Initialize instance properties."""
        self.inputs = inputs
        self.session_tc = session_tc

    @cached_property
    def tc(self) -> TC:
        """Return a case management instance."""
        return TC(self.inputs, self.session_tc)



"""TcEx Framework Module"""

# standard library
from typing import TYPE_CHECKING

# first-party
from tcex.api.tc.v2.threat_intelligence.mapping.indicator.indicator import Indicator

if TYPE_CHECKING:
    # first-party
    from tcex.api.tc.v2.threat_intelligence.threat_intelligence import ThreatIntelligence


class Address(Indicator):
    """Unique API calls for Address API Endpoints"""

    def __init__(self, ti: 'ThreatIntelligence', **kwargs):
        """Initialize instance properties.

        Args:
            ti (ThreatIntelligence): An instance of the ThreatIntelligence Class.
            ip (str): The value for this Indicator.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): [Read-Only] The date timestamp the Indicator was created.
            last_modified (str, kwargs): [Read-Only] The date timestamp the Indicator was last
                modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.
        """
        super().__init__(
            ti, sub_type='Address', api_entity='address', api_branch='addresses', **kwargs
        )
        self.unique_id = kwargs.get('unique_id', kwargs.get('ip'))
        self.data['ip'] = self.unique_id

    def _set_unique_id(self, json_response):
        """Set the unique_id provided a json response."""
        self.unique_id = json_response.get('ip', '')

    def can_create(self):
        """Return True if address can be created.

        If the ip address has been provided returns that the address can be created, otherwise
        returns that the address cannot be created.
        """
        return self.data.get('ip') is not None

    def dns_resolution(self):
        """Update the DNS resolution."""
        if not self.can_update():
            self._handle_error(910, [self.type])
        return self.tc_requests.dns_resolution(
            self.api_type, self.api_branch, self.unique_id, owner=self.owner
        )



"""TcEx Framework Module"""

# first-party
from tcex.input.model.api_model import ApiModel
from tcex.input.model.batch_model import BatchModel
from tcex.input.model.cert_model import CertModel
from tcex.input.model.logging_model import LoggingModel
from tcex.input.model.path_model import PathModel
from tcex.input.model.proxy_model import ProxyModel


class CommonModel(ApiModel, BatchModel, CertModel, LoggingModel, PathModel, ProxyModel):
    """Model Definition"""



"""TcEx Framework Module"""

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class OwnerModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='Owner Model',
    validate_assignment=True,
):
    """Owner Model"""

    _associated_type = PrivateAttr(default=False)
    _cm_type = PrivateAttr(default=False)
    _shared_type = PrivateAttr(default=False)
    _staged = PrivateAttr(default=False)

    id: int | None = Field(  # type: ignore
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    name: str | None = Field(
        None,
        allow_mutation=False,
        description='The name of the owner.',
        read_only=True,
        title='name',
    )
    owner_role: str | None = Field(
        None,
        allow_mutation=False,
        description="The user's role within the owner.",
        read_only=True,
        title='ownerRole',
    )
    perm_apps: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission is used for access to run/edit Apps.',
        read_only=True,
        title='permApps',
    )
    perm_artifact: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission used to access an Artifact.',
        read_only=True,
        title='permArtifact',
    )
    perm_attribute: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission is used for access to Attributes.',
        read_only=True,
        title='permAttribute',
    )
    perm_attribute_type: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission is used for access to Attribute Types.',
        read_only=True,
        title='permAttributeType',
    )
    perm_case_tag: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission used to access a Tag.',
        read_only=True,
        title='permCaseTag',
    )
    perm_comment: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission used to access a Comment.',
        read_only=True,
        title='permComment',
    )
    perm_copy_data: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission is used for access to Copy Data.',
        read_only=True,
        title='permCopyData',
    )
    perm_group: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission is used for access to Groups.',
        read_only=True,
        title='permGroup',
    )
    perm_indicator: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission is used for access to Indicators.',
        read_only=True,
        title='permIndicator',
    )
    perm_invite: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission is used for access to modify friends.',
        read_only=True,
        title='permInvite',
    )
    perm_members: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission is used for access to Members.',
        read_only=True,
        title='permMembers',
    )
    perm_playbooks: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission used with Playbooks.',
        read_only=True,
        title='permPlaybooks',
    )
    perm_playbooks_execute: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission used to execute Playbooks.',
        read_only=True,
        title='permPlaybooksExecute',
    )
    perm_post: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission is used for access to Posts.',
        read_only=True,
        title='permPost',
    )
    perm_publish: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission used to access Publications.',
        read_only=True,
        title='permPublish',
    )
    perm_security_label: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission is used for access to Security Labels.',
        read_only=True,
        title='permSecurityLabel',
    )
    perm_settings: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission is used for access to Organization Settings.',
        read_only=True,
        title='permSettings',
    )
    perm_tag: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission is used for access to Tags.',
        read_only=True,
        title='permTag',
    )
    perm_task: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission used to access a Task.',
        read_only=True,
        title='permTask',
    )
    perm_timeline: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission used to access a Timeline.',
        read_only=True,
        title='permTimeline',
    )
    perm_track: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission is used for access to Tracks.',
        read_only=True,
        title='permTrack',
    )
    perm_users: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission is used for access to User Settings.',
        read_only=True,
        title='permUsers',
    )
    perm_victim: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission is used for access to Victims.',
        read_only=True,
        title='permVictim',
    )
    perm_workflow_template: str | None = Field(
        None,
        allow_mutation=False,
        description='Permission used to access a Workflow Templates.',
        read_only=True,
        title='permWorkflowTemplate',
    )
    type: str | None = Field(
        None,
        allow_mutation=False,
        description='The owner type. Possible values: Organization, Community, Source.',
        read_only=True,
        title='type',
    )


class OwnerDataModel(
    BaseModel,
    title='Owner Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Owners Data Model"""

    data: list[OwnerModel] | None = Field(
        [],
        description='The data for the Owners.',
        methods=['POST', 'PUT'],
        title='data',
    )


class OwnersModel(
    BaseModel,
    title='Owners Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Owners Model"""

    _mode_support = PrivateAttr(default=False)

    data: list[OwnerModel] | None = Field(
        [],
        description='The data for the Owners.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# add forward references
OwnerDataModel.update_forward_refs()
OwnerModel.update_forward_refs()
OwnersModel.update_forward_refs()



"""TcEx Framework Module"""

# first-party
from tcex.app.service.api_service import ApiService
from tcex.app.service.common_service_trigger import CommonServiceTrigger
from tcex.app.service.mqtt_message_broker import MqttMessageBroker
from tcex.app.service.webhook_trigger_service import WebhookTriggerService

__all__ = ['ApiService', 'CommonServiceTrigger', 'MqttMessageBroker', 'WebhookTriggerService']
